"""Shell command related utils."""

from typing import Sequence
from pathlib import Path
import re
import subprocess as sp
import pandas as pd


def ls(path: Path | str) -> pd.DataFrame:
    """List files in the given path.
    This function is similar to the shell command `ls`.

    :param path: A path to a file or a directory.
    """
    if isinstance(path, str):
        path = Path(path)
    paths = list(path.iterdir()) if path.is_dir() else [path]
    return pd.DataFrame(
        {
            "paths": paths,
            "name": (p.name for p in paths),
        }
    )


def to_frame(
    cmd="",
    split: str = r"  +",
    header: int | list[str] | None = None,
    skip: int | Sequence[int] = (),
    lines: Sequence[str] = (),
    split_by_header: bool = False,
) -> pd.DataFrame:
    """Convert the result of a shell command to a DataFrame.
    The headers are splitted by a regular expression
    while the columns are splitted by the right-most position of the headers.

    :param cmd: A shell command.
    :param split: A regular expression pattern for splitting a line into fields.
    :param header: An integer, list of string or None specifiying the header of the data frame.
        If header is an integer,
        the corresponding row of lines after removing empty and skipped rows is used as header of the data frame;
        if header is a list of string then it is used as the header of the data frame.
        if header is None, then default header is used for the data frame.
    :param skip: Indexes of rows to skip.
    :param lines: The output of the shell command.
    :param split_by_header: If true, the headers are splitted by a regular expression
        and the columns are splitted by the right-most position of the headers.
        Otherwise, all lines are splitted by the specified regular expression.
    :return: A pandas DataFrame.
    """

    def _reg_skip(skip, n) -> set[int]:
        if isinstance(skip, int):
            skip = [skip]
        return set(idx % n for idx in skip)

    if not lines:
        lines = sp.check_output(cmd, shell=True).decode().strip().split("\n")
    skip = _reg_skip(skip, len(lines))
    lines = [line for idx, line in enumerate(lines) if idx not in skip]
    if split_by_header:
        return _to_frame_title(split=split, lines=lines)
    return _to_frame_space(split=split, header=header, lines=lines)


def _to_frame_space(
    lines: list[str],
    split: str = r"  +",
    header: int | list[str] | None = None,
) -> pd.DataFrame:
    """Convert the result of a shell command to a DataFrame.

    :param lines: The output of a shell command as lines of rows.
    :param split: A regular expression pattern for splitting a line into fields.
    :param header: An integer, list of string or None specifiying the header of the data frame.
        If header is an integer, the corresponding row of lines after removing empty and skipped rows is used as header of the data frame;
        if header is a list of string then it is used as the header of the data frame.
        if header is None, then default header is used for the data frame.
    :return: A pandas DataFrame.
    """
    data = [re.split(split, line.strip()) for line in lines if line.strip()]
    if isinstance(header, int):
        columns = [re.sub(r"\s+", "_", col.lower()) for col in data[header]]
        data = (row for idx, row in enumerate(data) if idx != header)
        frame = pd.DataFrame(data, columns=columns)
    elif isinstance(header, list):
        frame = pd.DataFrame(data, columns=header)
    else:
        frame = pd.DataFrame(data)
    return frame.astype(str)


def _to_frame_title(lines: list[str], split: str = r"  +") -> pd.DataFrame:
    """Convert the result of a shell command to a DataFrame.

    :param lines: The output of the shell command as list of lines.
    :param split: A regular expression pattern for splitting headers.
        Notice that non-header rows are splitted according the right-most position of the headers.
    :return: A pandas DataFrame.
    """
    headers = re.split(split, lines[0])
    n = len(headers)
    data = {}
    for idx in range(n - 1):
        start = lines[0].index(headers[idx])
        end = lines[0].index(headers[idx + 1])
        data[headers[idx]] = [line[start:end].strip() for line in lines[1:]]
    start = lines[0].index(headers[-1])
    data[headers[-1]] = [line[start:].strip() for line in lines[1:]]
    frame = pd.DataFrame(data)
    frame.columns = [col.strip().lower().replace(" ", "_") for col in frame.columns]
    return frame.astype(str)



"""Test dataframe.py."""

from pathlib import Path
import aiutil.dataframe

BASE_DIR = Path(__file__).resolve().parent


def test_read_csv():
    path = BASE_DIR / "data"
    df = aiutil.dataframe.read_csv(path)
    assert df.shape == (2, 2)



"""Script for fetch and filtering Spark application logs."""

from pathlib import Path
import re
from argparse import ArgumentParser, Namespace
import subprocess as sp
from loguru import logger
from .log import LogFilter

YARN = "/apache/hadoop/bin/yarn"


def filter_(args):
    """Filter the a log file.

    :param args: A Namespace object containing command-line options.
    """
    logf = LogFilter(
        log_file=args.log_file,
        context_size=args.context_size,
        keywords=args.keywords,
        patterns=args.patterns,
        output=args.output,
        threshold=args.threshold,
        dump_by_keyword=args.dump_by_keyword,
    )
    logf.filter()


def _format_app_id(app_id: str):
    app_id = app_id.lower()
    # support old Hive job id of the format job_123456789_123456
    app_id = re.sub("^job_", "application_", app_id)
    # support job id of the format _123456789_123456
    app_id = re.sub("^_", "application_", app_id)
    # support job id of the format 123456789_123456
    if not app_id.startswith("application_"):
        app_id = "application_" + app_id
    return app_id


def fetch(args):
    """Fetch and filter the log of a Spark/Hadoop application.

    :param args: A Namespace object containing command-line options.
    """
    app_id = _format_app_id(args.app_id)
    output = args.output if args.output else app_id
    cmd = [YARN, "logs", "-size_limit_mb", "-1", "-applicationId", app_id]
    if args.user:
        cmd = cmd + ["-appOwner", args.user]
    logger.info(f"Fetching log of the application {app_id} ...")
    with open(output, "w", encoding="utf-8") as fout:
        sp.run(cmd, stdout=fout, check=True)
    args.log_file = output
    filter_(args)


def status(args):
    """Get status of a Spark application.

    :param args: A Namespace object containing command-line options.
    """
    if args.app_id:
        cmd = ["yarn", "application", "-status", args.app_id]
        sp.run(cmd, check=True)
        return
    report = """Application Report : 
        Application-Id : {APP_ID}
        Application-Name : {APP_NAME}
        Application-Type : {APP_TYPE}
        User : {USER}
        Queue : {QUEUE}
        Application Priority : {PRIORITY}
        Start-Time : {START_TIME}
        Finish-Time : {FINISH_TIME}
        Progress : {PROGRESS}
        State : {STATE}
        Final-State : {STATUS}
        Tracking-URL : {URL}
        RPC Port : {PORT}
        AM Host : {HOST}
        Aggregate Resource Allocation : {RESOURCE}
        Log Aggregation Status : {LOG_STATUS}
        Diagnostics : 
        Unmanaged Application : {UNMANAGED}
        Application Node Label Expression : {APP_NODE_LABEL}
        AM container Node Label Expression : {CON_NODE_LABEL}
    """
    with args.log_file.open() as fin:
        for line in fin:
            if "{APP_ID}" in report:
                match = re.search(r"(application_\d+_\d+)", line)
                if match:
                    report = report.replace("{APP_ID}", match.group())
            if "{APP_NAME}" in report:
                match = re.search(r"--primary-py-file (.*) ", line)
                if match:
                    report = report.replace("{APP_NAME}", match.group())
            if "{USER}" in report:
                match = re.search(r"local/usercache/(.*)/", line)
                if match:
                    report = report.replace("{USER}", match.group())
            if "{STATUS}" in report:
                match = re.search(r"Final app status: (.*),", line)
                if match:
                    report = report.replace("{STATUS}", match.group())
    print(report)


def parse_args(args=None, namespace=None) -> Namespace:
    """Parse command-line arguments.

    :param args: The arguments to parse.
        If None, the arguments from command-line are parsed.
    :param namespace: An inital Namespace object.
    :return: A namespace object containing parsed options.
    """
    parser = ArgumentParser(description="Spark/Hadoop log utils.")
    subparsers = parser.add_subparsers(help="Sub commands.")
    _subparser_fetch(subparsers)
    _subparser_filter(subparsers)
    _subparser_status(subparsers)
    return parser.parse_args(args=args, namespace=namespace)


def _subparser_status(subparsers):
    subparser_status = subparsers.add_parser(
        "status", help="filter key information from a Spark/Hive application log."
    )
    mutex_group = subparser_status.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument(
        "-i", "--id", "--app-id", dest="app_id", type=str, help="An application ID."
    )
    mutex_group.add_argument(
        "-l", "-f", "--log-file", dest="log_file", type=Path, help="An application ID."
    )
    subparser_status.set_defaults(func=status)


def _option_filter(subparser) -> None:
    subparser.add_argument(
        "-k",
        "--keywords",
        nargs="+",
        dest="keywords",
        default=LogFilter.KEYWORDS,
        help="user-defined keywords to search for in the log file",
    )
    subparser.add_argument(
        "-i",
        "--ignore-patterns",
        nargs="+",
        dest="patterns",
        default=LogFilter.PATTERNS,
        help="regular expression patterns (date/time and ip by default) to ignore in dedup of filtered lines.",
    )
    subparser.add_argument(
        "-c",
        "--context-size",
        type=int,
        dest="context_size",
        default=3,
        help="number of lines (3 by default) to print before and after the suspicious line.",
    )
    subparser.add_argument(
        "-o",
        "--output",
        dest="output",
        default="",
        help="path of the output file (containing filtered lines).",
    )
    subparser.add_argument(
        "-t",
        "--threshold",
        dest="threshold",
        type=float,
        default=0.7,
        help="make pattern matching case-sensitive.",
    )
    subparser.add_argument(
        "-d",
        "--dump-by-keyword",
        dest="dump_by_keyword",
        action="store_true",
        help="dump error lines by keywords.",
    )


def _subparser_fetch(subparsers):
    subparser_fetch = subparsers.add_parser(
        "fetch", help="fetch the log of a Spark/Hive application."
    )
    subparser_fetch.add_argument(
        "app_id", help="the ID of the Spark/Hive application whose log is to fetch."
    )
    subparser_fetch.add_argument(
        "-u",
        "--user",
        dest="user",
        default=None,
        help="the name of the Spark/Hive application owner.",
    )
    subparser_fetch.add_argument(
        "-m",
        "--b-marketing-ep-infr",
        dest="user",
        action="store_const",
        const="b_marketing_ep_infr",
        help="Fetch log using the acount b_marketing_ep_infr.",
    )
    _option_filter(subparser_fetch)
    subparser_fetch.set_defaults(func=fetch)


def _subparser_filter(subparsers):
    subparser_filter = subparsers.add_parser(
        "filter", help="filter key information from a Spark/Hive application log."
    )
    subparser_filter.add_argument(
        "log_file", type=str, help="path of the log file to process"
    )
    _option_filter(subparser_filter)
    subparser_filter.set_defaults(func=filter_)


def main(args: Namespace | None = None):
    """The main function for script usage."""
    if args is None:
        args = parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# encoding: utf-8
"""A module makes it easy to run Scala/Python Spark job."""

from typing import Callable, Any, Iterable
import os
import sys
import itertools as it
from argparse import Namespace, ArgumentParser
from pathlib import Path
import tempfile
import shutil
import subprocess as sp
import re
import time
import datetime
import yaml
from loguru import logger
import notifiers
import aiutil.filesystem as fs


class SparkSubmit:
    """A class for submitting Spark jobs."""

    def __init__(self, email: dict | None = None, level: str = "INFO"):
        """Initialize a SparkSubmit instance.

        :param email: A dict object containing email information ("from", "to" and "host").
        :param level: The logging level for loguru.
        """
        # set up loguru with the right logging level
        try:
            logger.remove(0)
        except Exception:
            pass
        logger.add(sys.stdout, level=level)
        self._spark_submit_log = {}
        self.email = email

    def _spark_log_filter_helper_keyword(
        self,
        line: str,
        keyword: str,
        mutual_exclusive: list[str],
        time_delta: datetime.timedelta,
    ) -> bool:
        if keyword not in line:
            return False
        now = datetime.datetime.now()
        for kwd in mutual_exclusive:
            if kwd != keyword:
                self._spark_submit_log[kwd] = now - time_delta * 2
        if keyword not in self._spark_submit_log:
            self._spark_submit_log[keyword] = now
            return True
        if now - self._spark_submit_log[keyword] >= time_delta:
            self._spark_submit_log[keyword] = now
            return True
        return False

    def _spark_log_filter_helper_keywords(
        self,
        line: str,
        keywords: list[str],
        mutual_exclusive: bool,
        time_delta: datetime.timedelta,
    ) -> bool:
        mutual_exclusive = keywords if mutual_exclusive else ()
        for keyword in keywords:
            if self._spark_log_filter_helper_keyword(
                line=line,
                keyword=keyword,
                mutual_exclusive=mutual_exclusive,
                time_delta=time_delta,
            ):
                return True
        return False

    def _spark_log_filter(self, line: str) -> bool:
        line = line.strip().lower()
        if self._spark_log_filter_helper_keywords(
            line=line,
            keywords=["warn client", "uploading"],
            mutual_exclusive=False,
            time_delta=datetime.timedelta(seconds=0),
        ):
            return True
        if self._spark_log_filter_helper_keywords(
            line=line,
            keywords=["queue: ", "tracking url: "],
            mutual_exclusive=False,
            time_delta=datetime.timedelta(days=1),
        ):
            return True
        if self._spark_log_filter_helper_keywords(
            line=line,
            keywords=["exception", "user class threw", "caused by"],
            mutual_exclusive=False,
            time_delta=datetime.timedelta(seconds=1),
        ):
            return True
        if self._spark_log_filter_helper_keywords(
            line=line,
            keywords=["state: accepted", "state: running", "state: finished"],
            mutual_exclusive=True,
            time_delta=datetime.timedelta(minutes=10),
        ):
            return True
        if self._spark_log_filter_helper_keywords(
            line=line,
            keywords=[
                "final status: undefined",
                "final status: succeeded",
                "final status: failed",
            ],
            mutual_exclusive=True,
            time_delta=datetime.timedelta(minutes=3),
        ):
            return True
        return False

    @staticmethod
    def _filter(line: str, time_begin, log_filter: Callable | None = None) -> str:
        if not line:
            return ""
        if log_filter is None or log_filter(line):
            if "final status:" in line or " (state: " in line:
                line = line + f" (Time elapsed: {datetime.datetime.now() - time_begin})"
            return line
        return ""

    def submit(self, cmd: str, attachments: list[str] | None = None) -> bool:
        """Submit a Spark job.

        :param cmd: The Python script command to run.
        :param attachments: Attachments to send with the notification email.
        :return: True if the Spark application succeeds and False otherwise.
        """
        time_begin = datetime.datetime.now()
        logger.info("Submitting Spark job ...\n{}", cmd)
        stdout = []
        self._spark_submit_log.clear()
        with sp.Popen(cmd, shell=True, stderr=sp.PIPE) as process:
            while True:
                if process.poll() is None:
                    # pytype: disable=attribute-error
                    line = process.stderr.readline().decode().rstrip()
                    # pytype: enable=attribute-error
                    line = self._filter(line, time_begin, self._spark_log_filter)
                    if line:
                        print(line)
                        stdout.append(line)
                else:
                    for (
                        line
                    ) in process.stderr.readlines():  # pytype: disable=attribute-error
                        line = self._filter(
                            line.decode().rstrip(), time_begin, self._spark_log_filter
                        )
                        if line:
                            print(line)
                            stdout.append(line)
                    break
        # status
        status = self._final_status(stdout)
        app_id = self._app_id(stdout)
        if status:
            subject = f"Spark Application {app_id} {status}"
        else:
            subject = "Spark Application Submission Failed"
        if self.email:
            param = {
                "from_": self.email["from"],
                "to": self.email["to"],
                "subject": subject,
                "message": cmd + "\n".join(stdout),
                "host": self.email["host"],
                "username": "",
                "password": "",
            }
            if attachments:
                if isinstance(attachments, str):
                    attachments = [attachments]
                if not isinstance(attachments, list):
                    attachments = list(attachments)
                param["attachments"] = self._attach_txt(attachments)
            notifiers.get_notifier("email").notify(**param)
        if status == "FAILED":
            if self.email:
                self._notify_log(app_id, "Re: " + subject)
            return False
        return True

    @staticmethod
    def _attach_txt(attachments: list[str]) -> list[str]:
        dir_ = Path(tempfile.mkdtemp())
        paths = (dir_ / (Path(attach).name + ".txt") for attach in attachments)
        paths = [str(path) for path in paths]
        for attach, path in zip(attachments, paths):
            shutil.copy2(attach, path)
        return paths

    def _notify_log(self, app_id, subject):
        logger.info("Waiting for 300 seconds for the log to be available...")
        time.sleep(300)
        sp.run(f"logf fetch {app_id}", shell=True, check=True)
        lines = fs.filter(
            path=Path(app_id + "_s"),
            pattern=r"^-+\s+Deduped Error Lines\s+-+$",
            num_lines=999,
        )
        notifiers.get_notifier("email").notify(
            from_=self.email["from"],
            to=self.email["to"],
            subject="Re: " + subject,
            message="".join(lines[0]),
            host=self.email["host"],
            username="",
            password="",
        )

    @staticmethod
    def _app_id(stdout: list[str]) -> str:
        """Parse the application ID.

        :param stdout: Standard output as a list of strings.
        :return: The application ID of the Spark application.
        """
        for line in reversed(stdout):
            match = re.search(r"(application_\d+_\d+)", line)
            if match:
                return match.group(1)
        return ""

    @staticmethod
    def _final_status(stdout: list[str]) -> str:
        """Parse the final status of the Spark application.

        :param stdout: Standard output as a list of strings.
        :return: The final status (SUCCEED or FAILED) of the Spark application.
        """
        for line in reversed(stdout):
            match = re.search("final status: ([A-Z]+)", line)
            if match:
                return match.group(1)
        return ""


def _files(config: dict) -> str:
    """Get a list of valid configuration files to use with the option --files.

    :param config: A dict object containing configurations.
    :return: A string containing Spark configuration files separated by comma.
    """
    files = config["files"]
    files_xml = _files_xml(file for file in files if file.endswith(".xml"))
    files_non_xml = _files_non_xml(file for file in files if not file.endswith(".xml"))
    return ",".join(files_xml + files_non_xml)


def _file_exists(path: str) -> bool:
    if path.startswith("file://") and os.path.isfile(path[7:]):
        return True
    if path.startswith("viewfs://") or path.startswith("hdfs://"):
        process = sp.run(
            f"/apache/hadoop/bin/hdfs dfs -test -f {path}", shell=True, check=False
        )
        if process.returncode == 0:
            return True
    return False


def _get_first_valid_file(key: str, files: list[str]) -> str:
    for file in files:
        if _file_exists(file):
            return file
    logger.warning(
        "None of the specified configuration file for {} exists.\n    ",
        key,
        "\n".join("    " + file for file in files),
    )
    return ""


def _files_xml(files: Iterable[str]) -> list[str]:
    groups = [(key, list(val)) for key, val in it.groupby(files, os.path.basename)]
    files = (_get_first_valid_file(key, files) for key, files in groups)
    return [file for file in files if file]


def _files_non_xml(files: Iterable[str]) -> list[str]:
    res = []
    for file in files:
        if _file_exists(file):
            res.append(file)
        else:
            logger.warning("The file {} does NOT exist!", file)
    return res


def _python(config: dict) -> str:
    if "python-local" not in config:
        bins = ["python3", "python"]
    else:
        bins = config["python-local"]
        if isinstance(bins, str):
            bins = [bins]
    for bin_ in bins:
        if bin_.startswith("python"):
            bin_ = shutil.which(bin_)
        if bin_ and os.path.isfile(bin_):
            return bin_
    raise ValueError("No valid local python executable specified for python-local!")


def _submit_local(args, config: dict[str, Any]) -> bool:
    spark_submit = config.get("spark-submit-local", "")
    if not spark_submit:
        return True
    if not os.path.isfile(spark_submit):
        raise ValueError(f"{spark_submit} does not exist!")
    lines = [
        "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
        spark_submit,
    ]
    if config["jars"]:
        lines.append(f"--jars {config['jars']}")
    lines.extend(
        [
            "--conf spark.yarn.maxAppAttempts=1",
            "--conf spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT=1",
            "--conf spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT=1",
            "--conf spark.sql.execution.arrow.enabled=True",
        ]
    )
    python = _python(config)
    lines.append(f"--conf spark.pyspark.driver.python={python}")
    lines.append(f"--conf spark.pyspark.python={python}")
    lines.extend(args.pyfile)
    for idx in range(2, len(lines)):
        lines[idx] = " " * 4 + lines[idx]
    return SparkSubmit().submit(" \\\n".join(lines) + "\n", args.pyfile[:1])


def _submit_cluster(args, config: dict[str, Any]) -> bool:
    spark_submit = config.get("spark-submit", "")
    if not spark_submit:
        logger.warning("The filed spark-submit is not defined!")
        return True
    if not os.path.isfile(spark_submit):
        raise ValueError(f"{spark_submit} does not exist!")
    opts = (
        "files",
        "master",
        "deploy-mode",
        "queue",
        "num-executors",
        "executor-memory",
        "driver-memory",
        "executor-cores",
        "archives",
        "jars",
    )
    lines = (
        [config["spark-submit"]]
        + [f"--{opt} {config[opt]}" for opt in opts if opt in config and config[opt]]
        + [f"--conf {k}={v}" for k, v in config["conf"].items()]
    )
    lines.extend(args.pyfile)
    for idx in range(1, len(lines)):
        lines[idx] = " " * 4 + lines[idx]
    return SparkSubmit(email=config["email"]).submit(
        " \\\n".join(lines) + "\n", args.pyfile[:1]
    )


def submit(args: Namespace) -> None:
    """Submit the Spark job.

    :param args: A Namespace object containing command-line options.
    """
    # generate a config example
    if args.gen_config:
        path = Path(__file__).resolve().parent / "pyspark_submit.yaml"
        shutil.copy2(path, args.gen_config)
        logger.info("An example configuration is generated at {}", args.gen_config)
        return
    # load configuration
    if not args.config:
        config = {}
    else:
        with open(args.config, "r", encoding="utf-8") as fin:
            config = yaml.load(fin, Loader=yaml.FullLoader)
    # handle various options
    if args.spark_submit_local:
        config["spark-submit-local"] = args.spark_submit_local
    if args.python_local:
        config["python-local"] = args.python_local
    if "files" not in config:
        config["files"] = []
    config["files"].extend(args.files)
    config["files"] = _files(config)
    if "archives" in config:
        if isinstance(config["archives"], (list, tuple)):
            config["archives"] = ",".join(config["archives"])
    if "jars" in config:
        if isinstance(config["jars"], (list, tuple)):
            config["jars"] = ",".join(config["jars"])
    # submit Spark applications
    if _submit_local(args, config):
        _submit_cluster(args, config)


def parse_args(args=None, namespace=None) -> Namespace:
    """Parse command-line arguments.

    :param args: Arguments to parse.
    If None, arguments from command line is used.
    :param namespace: An initial Namespace object to use.
    :return: A Namespace object containing command-line options.
    """
    parser = ArgumentParser(description="Submit Spark application.")
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        default="",
        help="The configuration file to use.",
    )
    parser.add_argument(
        "--ssl",
        "--spark-submit-local",
        dest="spark_submit_local",
        default="",
        help="The local path to spark-submit.",
    )
    parser.add_argument(
        "--pl",
        "--python-local",
        dest="python_local",
        default="",
        help="The local path to Python.",
    )
    mutex_group = parser.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument(
        "-g",
        "--gen-config",
        "--generate-config",
        dest="gen_config",
        help="Specify a path for generating a configration example.",
    )
    mutex_group.add_argument(
        "--py",
        "--pyfile",
        dest="pyfile",
        nargs="+",
        help="The command (of PySpark script) to submit to Spark to run.",
    )
    parser.add_argument(
        "--files",
        dest="files",
        nargs="+",
        default=(),
        help="Additional files to upload.",
    )
    args = parser.parse_args(args=args, namespace=namespace)
    return args


def main():
    """Define a main function."""
    args = parse_args()
    submit(args)


if __name__ == "__main__":
    main()



"""Pandas DataFrame related utils."""

from pathlib import Path
import pandas as pd


def table_2w(
    frame: pd.DataFrame | pd.Series,
    columns: str | list[str] | None,
    na_as=None,
) -> pd.DataFrame:
    """Create 2-way table from columns of a DataFrame.

    :param frame: A pandas DataFrame.
    :param columns: Columns based on which to generate 2-way tables.
    :param na_as: The value to replace NAs.
    :raises TypeError: If frame is neither a pandas DataFrame nor a Series.
    :return: A 2-way table as a pandas DataFrame.
    """
    if na_as is not None:
        frame = frame.fillna(na_as)
    if isinstance(frame, pd.Series):
        df = frame.unstack()
        df.index = pd.MultiIndex.from_product([[df.index.name], df.index.values])
        df.columns = pd.MultiIndex.from_product([[df.columns.name], df.columns.values])
        return df
    if isinstance(frame, pd.DataFrame):
        if isinstance(columns, str):
            columns = [columns]
        return table_2w(frame[columns].groupby(columns).size(), columns=None)
    raise TypeError('"frame" must be pandas.Series or pandas.DataFrame.')


def read_csv(path: str | Path, **kwargs) -> pd.DataFrame:
    """Read many CSV files into a DataFrame at once.

    :param path: A path to a CSV file or to a directory containing CSV files.
    :param kwargs: Additional arguments to pass to pandas::read_csv.
    :return: A pandas DataFrame.
    """
    if isinstance(path, str):
        path = Path(path)
    if path.is_file():
        return pd.read_csv(path, **kwargs)
    return pd.concat(pd.read_csv(csv, **kwargs) for csv in path.glob("*.csv"))



"""Memory related utils."""

import getpass
import sys
import math
from collections import deque
import time
from argparse import ArgumentParser, Namespace
import numpy as np
import psutil
from loguru import logger

USER = getpass.getuser()


def get_memory_usage(user: str = USER) -> int:
    """Get the memory usage of the specified user.

    :param user: The user whose memory usage to get.
    :return: The memory usage of the user in bytes.
    """
    STATUS = (
        psutil.STATUS_RUNNING,
        psutil.STATUS_SLEEPING,
        psutil.STATUS_DISK_SLEEP,
        psutil.STATUS_WAKING,
        psutil.STATUS_PARKED,
        psutil.STATUS_IDLE,
        psutil.STATUS_WAITING,
    )
    try:
        return sum(
            p.memory_info().rss
            for p in psutil.process_iter()
            if p.username() == USER and p.status() in STATUS
        )
    except Exception:
        return get_memory_usage(user)


def monitor_memory_usage(seconds: float = 1, user: str = USER):
    """Log out the memory usage of the specified user in a specified frequency.

    :param seconds: The number of seconds to wait before the next logging.
    :param user: The user whose memory usage to monitor.
    """
    while True:
        time.sleep(seconds)
        logger.info("Memory used by {}: {:,}", user, get_memory_usage(user=user))


def match_memory_usage(
    target: float,
    arr_size: int = 1_000_000,
    sleep_min: float = 1,
    sleep_max: float = 30,
):
    """Match a user's memory usage to the specified value.
    The memory usage will gradually increase to the specified value
    if it is smaller than the specified value.
    Otherwise,
    the memory usage drops immediately to match the specified value.

    :param target: The target memory in bytes.
    :param arr_size: The size of integer arrays for consuming memory.
    :param sleep_min: The minimum time of sleeping.
    :param sleep_max: The maximum time of sleeping.
    """
    logger.info("Target memory: {:,.0f}", target)
    # define an template array
    arr = list(range(arr_size))
    size = sys.getsizeof(arr)
    # deque for consuming memory flexibly
    dq = deque()
    # define 2 points for linear interpolation of sleep seconds
    xp = (0, 10)
    yp = (sleep_max, sleep_min)
    while True:
        mem = get_memory_usage(USER)
        logger.info(
            "Current used memory by {}: {:,} out of which {:,} is contributed by the memory matcher",
            USER,
            mem,
            size * len(dq),
        )
        diff = (target - mem) / size
        if diff > 0:
            logger.info("Consuming more memory ...")
            dq.append(arr.copy())
            time.sleep(np.interp(diff, xp, yp))
        else:
            count = min(math.ceil(-diff), len(dq))
            logger.info("Releasing memory ...")
            for _ in range(count):
                dq.pop()
            time.sleep(np.interp(count, xp, yp))


def parse_args(args=None, namespace=None) -> Namespace:
    """Parse command-line arguments.

    :param args: The arguments to parse.
        If None, the arguments from command-line are parsed.
    :param namespace: An inital Namespace object.
    :return: A namespace object containing parsed options.
    """
    parser = ArgumentParser(
        description="Make memory consumption match the specified target."
    )
    mutex = parser.add_mutually_exclusive_group()
    mutex.add_argument(
        "-g",
        dest="target",
        type=lambda s: int(s) * 1073741824,
        help="Specify target memory in gigabytes.",
    )
    mutex.add_argument(
        "-m",
        dest="target",
        type=lambda s: int(s) * 1048576,
        help="Specify target memory in megabytes.",
    )
    return parser.parse_args(args=args, namespace=namespace)


def main():
    """The main function for scripting usage."""
    args = parse_args()
    match_memory_usage(args.target)


if __name__ == "__main__":
    main()



"""Easy development with uv managed Python projects."""

from typing import Iterable
from pathlib import Path
import os
import shutil
import subprocess as sp
import toml
from loguru import logger
import pathspec
from .filesystem import replace_patterns
from . import git

DIST = "dist"
README = "README.md"
TOML = "pyproject.toml"


def _project_dir() -> Path:
    """Get the root directory of the uv managed Python project.

    :return: The root directory of the uv managed Python project.
    :raises RuntimeError: Raises RuntimeError if the current directory is not under a uv managed Python project.
    """
    path = Path.cwd()
    while path.parent != path:
        if (path / TOML).is_file():
            return path
        path = path.parent
    return Path()


def _project_name(proj_dir: Path) -> str:
    """Get the name of the uv managed Python project.

    :param proj_dir: The root directory of the uv managed Python project.
    :return: The name of the project.
    """
    return toml.load(proj_dir / TOML)["project"]["name"]


def _project_version(proj_dir: Path) -> str:
    """Get the version of the uv managed Python project.

    :param proj_dir: The root directory of the uv managed Python project.
    :return: Version of the uv managed Python project.
    """
    return toml.load(proj_dir / TOML)["project"]["version"]


def _update_version_readme(ver: str, proj_dir: Path) -> None:
    """Update the version information in README.md.

    :param ver: The new version.
    :param proj_dir: The root directory of the  project.
    """
    replace_patterns(proj_dir / README, pattern=r"\d+\.\d+\.\d+", repl=f"{ver}")


def _update_version_toml(ver: str, proj_dir: Path) -> None:
    """Update the version information in the TOML file.

    :param ver: The new version.
    :param proj_dir: The root directory of the uv managed Python project.
    """
    replace_patterns(
        proj_dir / TOML, pattern=r"version = .\d+\.\d+\.\d+.", repl=f'version = "{ver}"'
    )


def _update_version_init(ver: str, proj_dir: Path) -> None:
    """Update the version information in the file __init__.py.

    :param ver: The new version.
    :param proj_dir: The root directory of uv managed Python project.
    """
    pkg = _project_name(proj_dir)
    for path in (proj_dir / pkg).glob("**/*.py"):
        replace_patterns(
            path,
            pattern=r"__version__ = .\d+\.\d+\.\d+.",
            repl=f'__version__ = "{ver}"',
        )


def _update_version(ver: str, proj_dir: Path) -> None:
    """Update versions in files.

    :param ver: The new version.
    :param proj_dir: The root directory of the uv managed Python project.
    """
    if ver:
        _update_version_init(ver=ver, proj_dir=proj_dir)
        _update_version_toml(ver, proj_dir=proj_dir)
        _update_version_readme(ver=ver, proj_dir=proj_dir)
        sp.run(["git", "diff"], check=True)


def version(
    ver: str = "",
    commit: bool = False,
    proj_dir: Path | None = None,
) -> None:
    """List or update the version of the uv managed Python package.

    :param ver: The new version to use.
        If empty, then the current version of the package is printed.
    :param commit: Whether to commit changes.
    :param proj_dir: The root directory of the uv managed Python project.
    """
    if proj_dir is None:
        proj_dir = _project_dir()
    if ver:
        _update_version(ver=ver, proj_dir=proj_dir)
        if commit:
            repo = git.Repo(root=proj_dir)
            repo.add()
            repo.commit(message="bump up version")
            repo.push()
    else:
        print(_project_version(proj_dir))


def add_tag_release(
    proj_dir: str | Path | None = None, tag: str = "", branch_release: str = "main"
) -> None:
    """Add a tag to the latest commit on the release branch for releasing.
    The tag is decided based on the current version of the project.

    :param proj_dir: The root directory of the uv managed Python project.
    :param tag: The tag (defaults to the current version of the package) to use.
    :param branch_release: The branch for releasing.
    :raises ValueError: If the tag to create already exists.
    """
    if proj_dir is None:
        proj_dir = _project_dir()
    if not tag:
        tag = "v" + _project_version(proj_dir)
    repo = git.Repo(proj_dir)
    if tag.encode() in repo.tag():
        raise ValueError(
            f"The tag {tag} already exists! Please merge new changes to the {branch_release} branch first."
        )
    branch_old = repo.active_branch()
    # add tag to the release branch
    repo.checkout(branch=branch_release)
    repo.pull(branch=branch_release)
    repo.tag(tag=tag)
    repo.push(branch=tag)
    # switch back to the old branch
    repo.checkout(branch=branch_old)


def format_code(
    commit: bool = False,
    proj_dir: Path | None = None,
    files: Iterable[Path | str] = (),
) -> None:
    """Format code.

    :param commit: If true (defaults to False),
        commit code formatting changes automatically.
    :param proj_dir: The root directory of the uv managed Python project.
    :param files: An iterable of Python scripts to format.
        If empty, then the whole project is formatted.
    """
    cmd = "uv run ruff format "
    if files:
        cmd += " ".join(f"'{file}'" for file in files)
    else:
        if proj_dir is None:
            proj_dir = _project_dir()
        cmd += str(proj_dir)
    logger.info("Formatting code using ruff ...")
    sp.run(cmd, shell=True, check=False, stdout=sp.PIPE)
    if commit:
        repo = git.Repo(proj_dir)
        repo.add()
        repo.commit(message="format code")
        repo.push()
        repo.status()


def _lint_code(proj_dir: Path | None, linter: str | list[str]):
    funcs = {
        "ruff": _lint_code_ruff,
        "pytype": _lint_code_pytype,
    }
    if isinstance(linter, str):
        linter = [linter]
    for lint in linter:
        funcs[lint](proj_dir)


def _lint_code_pytype(proj_dir: Path | None):
    logger.info("Linting code using pytype ...")
    if not proj_dir:
        proj_dir = _project_dir()
    pkg = _project_name(proj_dir)
    cmd = f"uv run pytype {proj_dir / pkg} {proj_dir / 'tests'}"
    try:
        sp.run(cmd, shell=True, check=True)
    except sp.CalledProcessError:
        logger.error("Please fix errors: {}", cmd)


def _lint_code_ruff(proj_dir: Path | None):
    logger.info("Linting code using ruff ...")
    if not proj_dir:
        proj_dir = _project_dir()
    pkg = _project_name(proj_dir)
    cmd = f"uv run ruff check {proj_dir / pkg}"
    try:
        sp.run(cmd, shell=True, check=True)
    except sp.CalledProcessError:
        logger.error("Please fix errors: {}", cmd)


def _lint_code_flake8(proj_dir: Path | None):
    logger.info("Linting code using flake8 ...")
    if not proj_dir:
        proj_dir = _project_dir()
    pkg = _project_name(proj_dir)
    cmd = f"PATH={proj_dir}/.venv/bin:$PATH flake8 {proj_dir / pkg}"
    try:
        sp.run(cmd, shell=True, check=True)
    except sp.CalledProcessError:
        logger.error("Please fix errors: {}", cmd)


def _lint_code_darglint(proj_dir: Path | None):
    logger.info("Linting docstring using darglint ...")
    if not proj_dir:
        proj_dir = _project_dir()
    pkg = _project_name(proj_dir)
    cmd = f"PATH={proj_dir}/.venv/bin:$PATH darglint {proj_dir / pkg}"
    try:
        sp.run(cmd, shell=True, check=True)
    except sp.CalledProcessError:
        logger.error("Please fix errors: {}", cmd)


def build_package(
    proj_dir: Path | None = None,
    linter: str | Iterable[str] = ("ruff", "pytype"),
    test: bool = True,
) -> None:
    """Build the package using uv.

    :param proj_dir: The root directory of the uv managed Python project.
    :param linter: A linter or an iterable of linters.
    :param test: Whether to run test suits (using pytest).
    :raises FileNotFoundError: If the command uv is not found.
    """
    if not shutil.which("uv"):
        raise FileNotFoundError("The command uv is not found!")
    if proj_dir is None:
        proj_dir = _project_dir()
    if os.path.exists(DIST):
        shutil.rmtree(DIST)
    _lint_code(proj_dir=proj_dir, linter=linter)
    format_code(proj_dir=proj_dir)
    if test:
        logger.info("Running unit tests...")
        sp.run(f"cd '{proj_dir}' && uv run pytest", shell=True, check=True)
    logger.info("Building the package...")
    sp.run(f"cd '{proj_dir}' && uv build", shell=True, check=True)


def clean(proj_dir: Path | None = None, ignore: str | Path | None = None) -> None:
    """Remove non-essential files from the current project.

    :param proj_dir: The root directory of the uv managed Python project.
    :param ignore: The full path to a GitIgnore file.
    """
    if proj_dir is None:
        proj_dir = _project_dir()
    if ignore is None:
        ignore = proj_dir / ".gitignore"
    elif isinstance(ignore, str):
        ignore = Path(ignore)
    if not ignore.is_file():
        return
    logger.info("Use the GitIgnore file: {}", ignore)
    with ignore.open("r", encoding="utf-8") as fin:
        patterns = [line.strip() for line in fin]
    spec = pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, patterns)
    _clean(proj_dir, spec)


def _clean(path: Path, spec: pathspec.PathSpec) -> None:
    if spec.match_file(path):
        if path.is_file():
            try:
                path.unlink()
            except Exception:
                logger.error("Failed to remove the file: {}", path)
        else:
            try:
                shutil.rmtree(path)
            except Exception:
                logger.error("Failed to remove the directory: {}", path)
        return
    if path.is_dir():
        for p in path.iterdir():
            _clean(p, spec)



"""Miscellaneous utils."""

from typing import Any, Sized, Callable
import time


def to_bool(value: Any) -> bool:
    """Convert an object to a bool value (True or False).

    :param value: any object that can be converted to a bool value.
    :return: True or False.
    """
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        if value.lower() in ("t", "true", "y", "yes"):
            return True
        if value.isdigit():
            return int(value) != 0
        return False
    if isinstance(value, int) and value != 0:
        return True
    if isinstance(value, Sized) and len(value) > 0:
        return True
    return False


def retry(task: Callable, times: int = 3, wait_seconds: float = 60):
    """Retry a Docker API on failure (for a few times).
    :param task: The task to run.
    :param times: The total number of times to retry.
    :param wait_seconds: The number of seconds to wait before retrying.
    :return: The return result of the task.
    """
    for _ in range(1, times):
        try:
            return task()
        except Exception:
            time.sleep(wait_seconds)
    return task()



#!/usr/bin/env python3
"""Search Jupyter/Lab notebooks."""

from typing import TypeAlias, Sequence
import json
from pathlib import Path
from collections import Counter
from argparse import ArgumentParser, Namespace
from loguru import logger

Criterion: TypeAlias = str | list[str] | dict[str, list[str]]


def _reg_criterion(criterion: str | list[str] | dict[str, list[str]]):
    """Convert different criterion specification into universal dict specification."""
    if isinstance(criterion, str):
        if criterion == "":
            criterion = []
        else:
            criterion = [criterion]
    if isinstance(criterion, list):
        criterion = {"include": criterion, "exclude": []}
    return criterion


class Cell:
    """Class representing a cell in a notebook."""

    def __init__(self, cell: dict):
        """Initialize a cell.
        :param cell: A cell in the dict representation.
        """
        self._cell = cell
        self._source = self.source(0)

    def source(self, indent: int = 0) -> str:
        """Get the source code in the cell.
        :param indent: The number of spaces to add before each line of code.
        """
        return "".join(" " * indent + line for line in self._cell["source"])

    def match_keyword(self, keyword: Criterion) -> bool:
        """Match source code of this cell against the specified keyword criterion.
        :param keyword: A keyword criterion.
        """
        kwd = _reg_criterion(keyword)
        return all(k in self._source for k in kwd["include"]) and not any(
            k in self._source for k in kwd["exclude"]
        )

    def match_type(self, type_: str) -> bool:
        """Match this cell's type against the specified one.
        :param type_: The type of cell to match.
        If empty, it matches all types.
        """
        if type_ == "":
            return True
        return self._cell["cell_type"] == type_


class Notebook:
    """A class representing a Jupyter/Lab notebook."""

    def __init__(self, path: str | Path):
        """Initialize a notebook.
        :param path: The path to a notebook.
        """
        self.path = Path(path) if isinstance(path, str) else path
        self._notebook = self._read_notebook()
        self.lang = self._get_lang().lower()
        self._cells = [Cell(cell) for cell in self._notebook["cells"]]

    def _get_lang(self) -> str:
        """Get the language of the notebook."""
        if "metadata" not in self._notebook:
            return "python"
        metadata = self._notebook["metadata"]
        if "kernelspec" in metadata:
            kernelspec = metadata["kernelspec"]
            if "language" in kernelspec:
                return kernelspec["language"]
            if "name" in kernelspec:
                return kernelspec["name"]
        elif "language_info" in metadata:
            return metadata["language_info"]["name"]
        return "python"

    def _read_notebook(self) -> dict:
        """Read the content of the notebook as a dict."""
        with self.path.open(encoding="utf-8") as fin:
            return json.load(fin)

    def match_language(self, language: Criterion) -> bool:
        """Match the language of the notebook against the specified language criterion.
        :param language: A language criterion.
        """
        lang = _reg_criterion(language)
        return all(self.lang == lg.lower() for lg in lang["include"]) and not any(
            self.lang == lg.lower() for lg in lang["exclude"]
        )

    def cells(self, keyword: Criterion, type_: str = "") -> list[Cell]:
        """Get a list of cells in this notebook."""
        return [
            cell
            for cell in self._cells
            if cell.match_type(type_) and cell.match_keyword(keyword)
        ]

    def __repr__(self) -> str:
        return f"Notebook({self.path})"


def print_nb_cells(
    nb_cells: tuple[tuple[Notebook, Cell], ...], num_notebooks: int, num_cells: int
):
    """Print cells of notebooks.
    :param nb_cells: Notebook cells to print.
    :param num_notebooks: The max number of notebooks to print.
    :param num_cells: The max number of cells per notebook to print.
    """
    n = len(nb_cells)
    print(f"Matched {n} notebooks")
    print(
        f"Display {min(n, num_notebooks)} notebooks each with up to {num_cells} cells\n"
    )
    for nb, cells in nb_cells[:num_notebooks]:
        print(f"{nb.path}: {nb.lang}")
        for idx, cell in enumerate(cells[:num_cells]):
            print(
                f"    ------------------------------------ Cell {idx} ------------------------------------"
            )
            print(cell.source(4))
        print(
            "========================================================================================\n\n"
        )


def search_notebooks(
    notebooks: list[Notebook],
    keyword: Criterion = "",
    type_: str = "",
    language: Criterion = "",
):
    """Search notebooks using the specified criterions.
    :param notebooks: A list of notebooks among which to search.
    :param keyword: A keyword criterion.
    :param type_: A type criterion.
    :param language: A language criterion.
    """
    notebooks = [nb for nb in notebooks if nb.match_language(language)]
    return tuple((nb, cells) for nb in notebooks if (cells := nb.cells(keyword, type_)))


def list_languages(notebooks: list[Notebook]) -> list[tuple[str, int]]:
    """List languages used in notebooks.
    :param notebooks: A list of notebooks.
    """
    counter = Counter(nb.lang for nb in notebooks)
    counter = list(counter.items())
    counter.sort(key=lambda t: -t[1])
    return counter


def find_notebooks(paths: Sequence[str]) -> list[Notebook]:
    """Find all notebooks under the given paths.
    :param paths: A list of paths.
    """
    notebooks = set()
    for path in paths:
        path = Path(path)
        if path.is_file():
            if path.suffix == ".ipynb":
                notebooks.add(path)
            else:
                logger.warning(f"The file {path} is not a notebook!")
        elif path.is_dir():
            for p in path.glob("**/*.ipynb"):
                notebooks.add(p)
    return [Notebook(path) for path in notebooks]


def _list_langs_args(args):
    notebooks = find_notebooks(args.paths)
    counter = list_languages(notebooks)
    for lang, freq in counter:
        print(f"{lang}: {freq}")
    print()


def _search_notebooks_args(args):
    notebooks = find_notebooks(args.paths)
    lang = {
        "include": args.lang_include,
        "exclude": args.lang_exclude,
    }
    kwd = {
        "include": args.kwd_include,
        "exclude": args.kwd_exclude,
    }
    nb_cells = search_notebooks(notebooks, keyword=kwd, language=lang)
    print_nb_cells(nb_cells, args.num_notebooks, args.num_cells)


def parse_args(args=None, namespace=None) -> Namespace:
    """Parse command-line arguments.
    :param args: The arguments to parse.
        If None, the arguments from command-line are parsed.
    :param namespace: An inital Namespace object.
    """
    parser = ArgumentParser(description="Search for notebooks.")
    subparsers = parser.add_subparsers(dest="sub_cmd", help="Sub commands.")
    _subparse_search(subparsers)
    _subparse_list(subparsers)
    return parser.parse_args(args=args, namespace=namespace)


def _subparse_list(subparsers):
    subparser_list = subparsers.add_parser(
        "list",
        aliases=["l", "ls"],
        help="List languages used by notebooks.",
    )
    subparser_list.add_argument(
        "-p",
        "--paths",
        dest="paths",
        nargs="+",
        required=True,
        help="Paths to notebooks or directories containing notebooks.",
    )
    subparser_list.set_defaults(func=_list_langs_args)


def _subparse_search(subparsers):
    subparser_search = subparsers.add_parser(
        "search",
        aliases=["s"],
        help="Search for notebooks.",
    )
    subparser_search.add_argument(
        "-p",
        "--paths",
        dest="paths",
        nargs="+",
        required=True,
        help="Paths to notebooks or directories containing notebooks.",
    )
    subparser_search.add_argument(
        "-l",
        "--lang-include",
        dest="lang_include",
        nargs="*",
        default=(),
        help="The language of notebooks.",
    )
    subparser_search.add_argument(
        "-L",
        "--lang-exclude",
        dest="lang_exclude",
        nargs="*",
        default=(),
        help="Languages that notebooks shouldn't include.",
    )
    subparser_search.add_argument(
        "-k",
        "--kwd-include",
        dest="kwd_include",
        nargs="*",
        default=(),
        help="Keywords to search for in cells of notebooks.",
    )
    subparser_search.add_argument(
        "-K",
        "--kwd-exclude",
        dest="kwd_exclude",
        nargs="*",
        default=(),
        help="Keywords that cells of notebooks shouldn't include.",
    )
    subparser_search.add_argument(
        "-n",
        "--num-notebooks",
        dest="num_notebooks",
        type=int,
        default=10,
        help="Number of matched notebooks to display.",
    )
    subparser_search.add_argument(
        "-c",
        "--num-cells",
        dest="num_cells",
        type=int,
        default=10,
        help="Number of matched cells in each notebook to display.",
    )
    subparser_search.set_defaults(func=_search_notebooks_args)


def main() -> None:
    """The main function of the script."""
    args = parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



"""Manipulating PDFs."""

from typing import Iterable
from pypdf import PdfWriter, PdfReader


def extract_pages(file: str, subfiles: dict[str, int | Iterable[int]]) -> None:
    """Extract pages from a PDF file and write into sub PDF file.

    :param file: The raw PDF file to extract pages from.
    :param subfiles: A dictionary specifying sub PDF files
        and their corresponding indexes (0-based) of pages from the raw PDF file.
        For example,
        the following code extract pages 0-4 as first.pdf,
        pages 5 and 7 as second.pdf,
        and page 6 as third.pdf from raw.pdf.
        .. highlight:: python
        .. code-block:: python

        from aiutil.pdf import extract_pages
        extract_pages("raw.pdf", {"first.pdf": range(5), "second.pdf": [5, 7], "third.pdf": 6})
    """
    with open(file, "rb") as fin:
        reader = PdfReader(fin)
        for subfile, indexes in subfiles.items():
            _extract_pages(reader, indexes, subfile)


def _extract_pages(
    reader: PdfReader, indexes: int | Iterable[int], output: str
) -> None:
    """A helper function for extract_pages.

    :param reader: A PdfFileReader object.
    :param indexes: Index (0-based) of pages to extract.
    :param output: The path of the sub PDF file to write the extracted pages to.
    """
    writer = PdfWriter()
    if isinstance(indexes, int):
        indexes = [indexes]
    for index in indexes:
        writer.add_page(reader.pages[index])
    with open(output, "wb") as fout:
        writer.write(fout)



"""URL related utility functions."""


def format(url: str, text: str = "", type_="html") -> str:
    """Convert an URL to be a HTML hyperlink or a hyperlink formula in Excel so that it can be opened by clicking.

    :param url: The url to be converted to a hyperlink.
    :param text: The text to display.
    :param type_: The type (html or excel) of hyperlink.
    :return: The formatted URL.
    """
    if not text:
        text = url
    type_ = type_.strip().lower()
    if type_ == "html":
        return f'<a href="{url}" target="_blank"> {text} </a>'
    if type_ == "excel":
        return f'=HYPERLINK("{url}", "{text}")'
    return url



"""Test the aiutil.jupyter module."""

from pathlib import Path
import aiutil.notebook.util

BASE_DIR = Path(__file__).parent


def test_nbconvert_notebooks():
    aiutil.notebook.util.nbconvert_notebooks(BASE_DIR / "notebooks")



"""Computer vision related utils."""

from typing import Iterable
from pathlib import Path
from tqdm import tqdm, trange
import numpy as np
import pandas as pd
from PIL import Image
import skimage
import cv2


def video_to_image(
    file: str,
    step: int,
    bbox: tuple[int, int, int, int] | None = None,
    output: str = "frame_{:0>7}.png",
):
    """Extract images from a video file.

    :param file: The path to video file.
    :param bbox: A bounding box.
        If specified, crop images using the bounding box.
    :param output_dir: The directory to save extracted images.
    :param step: Keep 1 image every step frames.
    """
    Path(output.format(0)).parent.mkdir(parents=True, exist_ok=True)
    vidcap = cv2.VideoCapture(file)
    total = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    for idx in trange(total):
        success, arr = vidcap.read()
        if not success:
            break
        if idx % step == 0:
            img = Image.fromarray(np.flip(arr, 2))
            if bbox:
                img = img.crop(bbox)
            img.save(output.format(idx))
    vidcap.release()


def resize_image(
    paths: str | Path | Iterable[Path],
    desdir: str | Path | None,
    size: tuple[int, int],
) -> None:
    """Resize images to a given size.

    :param paths: The paths to images to be resized.
    :param desdir: The directory to save resized images.
        Notice that both '.' and '""' stand for the current directory.
        If None is specified, then the orginal image is overwritten.
    :param size: The new size of images.
    """
    if isinstance(desdir, str):
        desdir = Path(desdir)
    if isinstance(desdir, Path):
        desdir.mkdir(exist_ok=True)
    if isinstance(paths, str):
        paths = Path(paths)
    if isinstance(paths, Path):
        img = Image.open(paths)
        if img.size != size:
            img.resize(size).save(desdir / paths.name if desdir else paths)
        return
    if not hasattr(paths, "__len__"):
        paths = tuple(paths)
    for path in tqdm(paths):
        resize_image(paths=path, desdir=desdir, size=size)


def _is_approx_close(x: float, y: float, threshold: float = 0.4) -> bool:
    """Helper function of is_approx_close.
    Check whether the 2 values x and y are relative close.

    :param x: An non negative value.
    :param y: Another non negative value.
    :param threshold: The maximum ratio difference from 1 to be considered as close.
    :return: True if the 2 values are considered close and False otherwise.
    """
    if x < y:
        x, y = y, x
    return (x + 0.01) / (y + 0.01) <= 1 + threshold


def is_approx_close(red: int, green: int, blue: int, threshold: float = 0.4) -> bool:
    """Check whether the 3 channels have approximately close values.

    :param red: The red channel.
    :param green: The green channel.
    :param blue: The blue channel.
    :param threshold: The threshold (absolute deviation from 1)
        to consider a ratio (of 2 channels) to be close to 1.
    :return: True if the RGB values are approximately close to each other.
    """
    return (
        _is_approx_close(red, green, threshold=threshold)
        and _is_approx_close(red, blue, threshold=threshold)
        and _is_approx_close(green, blue, threshold=threshold)
    )


def deshade_arr_1(arr: np.ndarray, threshold: float = 0.4) -> np.ndarray:
    """Deshade a poker card (i.e., get rid of the shading effec on a poker card)
        by checking whether the 3 channels have relative close values.

    :param arr: A numpy.ndarray representation of the image to be deshaded.
    :param threshold: The threshold (absolute deviation from 1)
        to consider a ratio (of 2 channels) to be close to 1.
    :return: A new numpy ndarray with shading effect removed.
    """
    arr = arr.copy()
    nrow, ncol, _ = arr.shape
    for i in range(nrow):
        for j in range(ncol):
            r = arr[i, j, 0]
            g = arr[i, j, 1]
            b = arr[i, j, 2]
            if is_approx_close(r, g, b, threshold=threshold):
                arr[i, j, :] = (255, 255, 255)
    return arr


def deshade_arr_2(arr: np.ndarray, cutoff: float = 30) -> np.ndarray:
    """Deshade a poker card (i.e., get rid of the shading effec on a poker card)
        by checking whether the 3 channels all have values larger than a threshold.

    :param arr: A numpy.ndarray representation of the image to be deshaded.
    :param cutoff: The cutoff value of 3 channels.
        If the 3 channels all have value no less than this cutoff,
        then it is considered as shading effect.
    :return: A new numpy ndarray with shading effect removed.
    """
    arr = arr.copy()
    nrow, ncol, _ = arr.shape
    for i in range(nrow):
        for j in range(ncol):
            r = arr[i, j, 0]
            g = arr[i, j, 1]
            b = arr[i, j, 2]
            # if (r + g + b) / 3 >= cutoff and max(r, g, b) <= 150:
            if min(r, g, b) >= cutoff:
                arr[i, j, :] = (255, 255, 255)
    return arr


def deshade_arr_3(
    arr: np.ndarray, threshold: float = 0.4, cutoff: float = 30
) -> np.ndarray:
    """Deshade a poker card (i.e., get rid of the shading effect on a poker card)
        by combining methods in deshade_arr_1 and deshade_arr_2.

    :param arr: A numpy.ndarray representation of the image to be deshaded.
    :param threshold: The threshold (absolute deviation from 1)
        to consider a ratio (of 2 channels) to be close to 1.
    :param cutoff: The cutoff value of 3 channels.
        If the 3 channels all have value no less than this cutoff,
        then it is considered as shading effect.
    :return: A new numpy ndarray with shading effect removed.
    """
    arr = arr.copy()
    nrow, ncol, _ = arr.shape
    for i in range(nrow):
        for j in range(ncol):
            r = arr[i, j, 0]
            g = arr[i, j, 1]
            b = arr[i, j, 2]
            if min(r, g, b) >= cutoff and is_approx_close(r, g, b, threshold=threshold):
                arr[i, j, :] = (255, 255, 255)
    return arr


def deshade_1(img, threshold=0.4) -> Image.Image:
    """Deshade an image (i.e., get rid of the shading effec on an image.)
        by checking whether the 3 channels have relative close values.

    :param img: An image to deshade.
    :param threshold: The threshold (absolute deviation from 1)
        to consider a ratio (of 2 channels) to be close to 1.
    :return: The new image with shading effect removed.
    """
    arr = np.array(img)
    arr = deshade_arr_1(arr, threshold=threshold)
    return Image.fromarray(arr)


def deshade_2(img, cutoff=30) -> Image.Image:
    """Deshade an image (i.e., get rid of the shading effec on an image)
        by checking whether the 3 channels all have values larger than a threshold.

    :param img: An image to deshade.
    :param cutoff: The cutoff value of 3 channels.
        If the 3 channels all have value no less than this cutoff,
        then it is considered as shading effect.
    :return: The new image with shading effect removed.
    """
    arr = np.array(img)
    arr = deshade_arr_2(arr, cutoff=cutoff)
    return Image.fromarray(arr)


def deshade_3(img, threshold=0.4, cutoff=30) -> Image.Image:
    """Deshade an image (i.e., get rid of the shading effect on an image)
        by combining methods in deshade_arr_1 and deshade_arr_2.

    :param img: An image to deshade.
    :param threshold: The threshold (absolute deviation from 1)
        to consider a ratio (of 2 channels) to be close to 1.
    :param cutoff: The cutoff value of 3 channels.
        If the 3 channels all have value no less than this cutoff,
        then it is considered as shading effect.
    :return: The new image with shading effect removed.
    """
    arr = np.array(img)
    arr = deshade_arr_3(arr, threshold=threshold, cutoff=cutoff)
    return Image.fromarray(arr)


def add_frames(
    arr: np.ndarray | Image.Image,
    bboxes: list[tuple[int, int, int, int]],
    rgb: tuple[int, int, int] = (255, 0, 0),
) -> np.ndarray:
    """Add (highlighting) frames into an image.

    :param arr: A PIL image or its numpy array representation.
    :param bboxes: A list of bounding boxes.
    :param rgb: The RGB color (defaults to (255, 0, 0)) of the (highlighting) frame.
    :return: A numpy array representation of the altered image.
    """
    if isinstance(arr, Image.Image):
        arr = np.array(arr)
    for x1, y1, x2, y2 in bboxes:
        if x2 is None:
            x2 = x1
        if y2 is None:
            y2 = y1
        arr[y1, x1:x2, :] = rgb
        arr[y2, x1:x2, :] = rgb
        arr[y1:y2, x1, :] = rgb
        arr[y1:y2, x2, :] = rgb
    return arr


def duplicate_image(
    path: str | Path,
    copies: int,
    des_dir: str | Path | None = None,
    noise_amount: float = 0.05,
):
    """Duplicate an image with some noises added.

    :param path: The path to the image to be duplicated.
    :param copies: The number of copies to duplicate.
    :param noise_amount: Proportion of image pixels to replace with noise on range [0, 1].
    """
    if isinstance(path, str):
        path = Path(path)
    if isinstance(des_dir, str):
        des_dir = Path(des_dir)
    if des_dir is None:
        des_dir = path.parent
    des_dir.mkdir(parents=True, exist_ok=True)
    for i in range(copies):
        file_i = des_dir / f"{path.stem}_copy{i}.png"
        noise = skimage.util.random_noise(
            np.array(Image.open(path)), mode="s&p", amount=noise_amount
        )
        Image.fromarray(np.array(noise * 255, dtype=np.uint8)).save(file_i)


def structural_similarity(im1, im2) -> float:
    """Extend skimage.metrics.structural_similarity
        to calculate the similarity of (any) two images.

    :param im1: A PIL image.
    :param im2: Another PIL image.
    """
    size = im1.size
    if im2.size != size:
        im2 = im2.resize(size)
    return skimage.metrics.structural_similarity(
        np.array(im1), np.array(im2), multichannel=True
    )


def calc_image_similarities(img: Image.Image | str | Path, dir_: str | Path):
    """Calculate the similarities between an image and all images in a directory.

    :param img: A PIL image or the path to an image file.
    :param dir_: A directory containing images.
    """
    if isinstance(img, (str, Path)):
        img = Image.open(img)
    if isinstance(dir_, str):
        dir_ = Path(dir_)
    paths = list(dir_.glob("*.png"))
    sims = [structural_similarity(img, Image.open(p)) for p in tqdm(paths)]
    return pd.DataFrame(
        {
            "path": paths,
            "similarity": sims,
        }
    )



"""SQL related utils."""

from pathlib import Path
import subprocess as sp
import sqlparse


def format(path: Path | str):
    """Format a SQL file.

    :param path: The path to a SQL file.
    """
    if isinstance(path, str):
        path = Path(path)
    query = sqlparse.format(
        path.read_text(encoding="utf-8"),
        keyword_case="upper",
        identifier_case="lower",
        strip_comments=False,
        reindent=True,
        indent_width=2,
    )
    path.write_text(query, encoding="utf-8")
    cmd = f"pg_format --function-case 1 --type-case 3 --inplace {path}"
    sp.run(cmd, shell=True, check=True)



"""Test dataframe.py."""

from pathlib import Path
import aiutil.filesystem

BASE_DIR = Path(__file__).resolve().parent


def test_is_ess_empty():
    assert aiutil.filesystem.is_ess_empty(BASE_DIR) is False
    assert aiutil.filesystem.is_ess_empty(BASE_DIR / "ess_empty")
    assert aiutil.filesystem.is_ess_empty(BASE_DIR / "ess_empty/.ipynb_checkpoints")



import sys
from pathlib import Path
import pytest

BASE_DIR = Path(__file__).parent


@pytest.mark.skipif(sys.platform == "win32", reason="Skip test on Windows")
def test_shebang():
    import aiutil.shebang

    aiutil.shebang.update_shebang(BASE_DIR / "script_dir", "#!/usr/bin/env python3")



"""PYPI related utils."""

from typing import Set
import re
import requests


def dep(pkg: str, recursive: bool = False) -> set[str]:
    """Parse dependencies of the specified Python package.

    :param pkg: The Python package whose dependencies to parse.
    :param recursive: If true, pass dependencies recursively.
    :return: A set of Python package names
        which are dependencies of the specified Python package.
    """
    if recursive:
        return _dep_recur(pkg)
    return _dep(pkg)


def _dep(pkg):
    url = f"https://pypi.org/pypi/{pkg}/json"
    deps = requests.get(url, timeout=10).json()["info"]["requires_dist"]
    if deps is None:
        return set()
    return set(dep for dep in deps if "extra ==" not in dep)


def _dep_recur(pkg: str):
    deps = set()
    _dep_recur_helper(pkg, deps)
    return deps


def _dep_recur_helper(pkg: str, deps: Set[str]):
    for dep in _dep(pkg):
        dep = re.split(r" |;|\(|\[", dep, maxsplit=1)[0]
        if dep not in deps:
            _dep_recur_helper(dep, deps)
            deps.add(dep)



#!/usr/bin/env python3
"""Filesystem related util functions."""

from collections import namedtuple
from typing import Iterable, Callable
import itertools
import os
import sys
import re
import shutil
import math
from pathlib import Path
import subprocess as sp
from itertools import chain
import tempfile
from tqdm import tqdm
import pandas as pd
from loguru import logger
import dulwich.porcelain

HOME = Path.home()
PosixPathPair = namedtuple("PosixPathPair", ["prefix", "base"])


def copy_if_exists(src: str, dst: str | Path = HOME) -> bool:
    """Copy a file.
    No exception is thrown if the source file does not exist.

    :param src: The path of the source file.
    :param dst: The path of the destination file.
    :return: True if a copy if made, vice versa.
    """
    if not os.path.exists(src):
        return False
    try:
        shutil.copy2(src, dst)
        return True
    except Exception:
        return False


def link_if_exists(
    src: str, dst: str | Path = HOME, target_is_directory: bool = True
) -> bool:
    """Make a symbolic link of a file.
    No exception is thrown if the source file does not exist.

    :param src: The path of the source file.
    :param dst: The path of the destination file.
    :param target_is_directory: Whether the target is a directory.
    :return: True if a symbolic link is created, vice versa.
    """
    if not os.path.exists(src):
        return False
    if os.path.exists(dst):
        shutil.rmtree(dst)
    try:
        os.symlink(src, dst, target_is_directory=target_is_directory)
        return True
    except Exception:
        return False


def count_path(
    paths: Iterable[str],
    weights: Iterable[int | float] | None = None,
) -> pd.Series:
    """Count frequence of paths and their parent paths.

    :param paths: An iterable collection of paths.
    :param weights: Weights of paths.
    :return: A pandas Series with paths as index and frequencies of paths as value.
    """

    def _count_path_helper(path: str, weight: int | float, freq: dict) -> None:
        fields = path.rstrip("/").split("/")[:-1]
        path = ""
        for field in fields:
            path = path + field + "/"
            freq[path] = freq.get(path, 0) + weight

    freq = {}
    if weights is None:
        weights = itertools.repeat(1)
    for path, weight in zip(paths, weights):
        _count_path_helper(path, weight, freq)
    return pd.Series(freq, name="count")


def zip_subdirs(root: str | Path) -> None:
    """Compress subdirectories into zip files.

    :param root: The root directory whose subdirs are to be zipped.
    """
    if isinstance(root, str):
        root = Path(root)
    for path in root.iterdir():
        if path.is_dir() and not path.name.startswith("."):
            file = path.with_suffix(".zip")
            print(f"{path} -> {file}")
            sp.run(f"zip -qr {file} {path} &", shell=True, check=True)


def flatten_dir(dir_: str | Path) -> None:
    """Flatten a directory,
    i.e., move files in immediate subdirectories into the current directory.

    :param dir_: The directory to flatten.
    """
    if isinstance(dir_, str):
        dir_ = Path(dir_)
    for path in dir_.iterdir():
        if path.is_dir():
            _flatten_dir(path)
            path.rmdir()


def _flatten_dir(dir_: Path) -> None:
    """Helper method of flatten_dir.

    :param dir_: A directory to flatten.
    """
    for path in dir_.iterdir():
        path.rename(path.parent.parent / path.name)


def split_dir(dir_: str | Path, batch_size: int, wildcard: str = "*") -> None:
    """Split files in a directory into sub-directories.
    This function is for the convenience of splitting a directory
    with a large number of files into smaller directories
    so that those subdirs can zipped (into relatively smaller files) and uploaded to cloud quickly.

    :param dir_: The root directory whose files are to be splitted into sub-directories.
    :param wildcard: A wild card pattern specifying files to be included.
    :param batch_size: The number files that each subdirs should contain.
    """
    if isinstance(dir_, str):
        dir_ = Path(dir_)
    files = sorted(dir_.glob(wildcard))
    num_batch = math.ceil(len(files) / batch_size)
    nchar = len(str(num_batch))
    for batch_idx in tqdm(range(num_batch)):
        _split_dir_1(dir_ / f"{batch_idx:0>{nchar}}", files, batch_idx, batch_size)


def _split_dir_1(
    desdir: Path, files: list[Path], batch_idx: int, batch_size: int
) -> None:
    """Helper method of split_dir.

    :param desdir: The destination directory to save subset of files.
    :param files: A list of Path objects.
    :param batch_idx: The batch index (0-based).
    :param batch_size: The size of the batch.
    """
    desdir.mkdir(exist_ok=True)
    for path in files[(batch_idx * batch_size) : ((batch_idx + 1) * batch_size)]:
        path.rename(desdir / path.name)


def find_images(root_dir: str | Path | list[str | Path]) -> list[Path]:
    """Find all PNG images in a (sequence) of dir(s) or its/their subdirs.

    :param root_dir: A (list) of dir(s).
    :return: A list of Path objects to PNG images.
    """
    if isinstance(root_dir, (str, Path)):
        root_dir = [root_dir]
    images = []
    for path in root_dir:
        if isinstance(path, str):
            path = Path(path)
        images.extend(path.glob("**.png"))
    return images


def find_data_tables(
    root: str | Path,
    filter_: Callable = lambda _: True,
    extensions: Iterable[str] = (),
    patterns: Iterable[str] = (),
) -> set[str]:
    """Find keywords which are likely data table names.

    :param root: The root directory or a GitHub repo URL in which to find data table names.
    :param filter_: A function for filtering identified keywords (via regular expressions).
        By default, all keywords identified by regular expressions are kept.
    :param extensions: Addtional text file extensions to use.
    :param patterns: Addtional regular expression patterns to use.
    :return: A set of names of data tables.
    """
    if isinstance(root, str):
        if re.search(r"(git@|https://).*\.git", root):
            with tempfile.TemporaryDirectory() as tempdir:
                dulwich.porcelain.clone(root, tempdir)
                logger.info(
                    "The repo {} is cloned to the local directory {}.", root, tempdir
                )
                return find_data_tables(tempdir, filter_=filter_)
        root = Path(root)
    if root.is_file():
        return _find_data_tables_file(root, filter_, patterns)
    extensions = {
        ".sql",
        ".py",
        ".ipy",
        ".ipynb",
        ".scala",
        ".java",
        ".txt",
        ".json",
    } | set(extensions)
    paths = (
        path
        for path in Path(root).glob("**/*")
        if path.suffix.lower() in extensions and path.is_file()
    )
    return set(
        chain.from_iterable(
            _find_data_tables_file(path, filter_, patterns) for path in paths
        )
    )


def _find_data_tables_file(file, filter_, patterns) -> set[str]:
    if isinstance(file, str):
        file = Path(file)
    text = file.read_text(encoding="utf-8").lower()
    patterns = {
        r"from\s+(\w+)\W*\s*",
        r"from\s+(\w+\.\w+)\W*\s*",
        r"join\s+(\w+)\W*\s*",
        r"join\s+(\w+\.\w+)\W*\s*",
        r"table\((\w+)\)",
        r"table\((\w+\.\w+)\)",
        r'"table":\s*"(\w+)"',
        r'"table":\s*"(\w+\.\w+)"',
    } | set(patterns)
    tables = chain.from_iterable(re.findall(pattern, text) for pattern in patterns)
    mapping = str.maketrans("", "", "'\"\\")
    tables = (table.translate(mapping) for table in tables)
    return set(table for table in tables if filter_(table))


def find_data_tables_sql(sql: str, filter_: Callable | None = None) -> set[str]:
    """Find keywords which are likely data table names in a SQL string.

    :param sql: A SQL query.
    :param filter_: A function for filtering identified keywords (via regular expressions).
        By default, all keywords identified by regular expressions are kept.
    :return: A set of names of data tables.
    """
    sql = sql.lower()
    pattern = r"(join|from)\s+(\w+(\.\w+)?)(\s|$)"
    tables = (pms[1] for pms in re.findall(pattern, sql))
    if filter_ is None:
        return set(tables)
    return set(table for table in tables if filter_(table))


def is_empty(dir_: str | Path, filter_: Callable | None = lambda _: True) -> bool:
    """Check whether a directory is empty.

    :param dir_: The directory to check.
    :param filter_: A filtering function (default True always) to limit the check to sub files/dirs.
    :return: True if the specified directory is empty and False otherwise.
    """
    if isinstance(dir_, str):
        dir_ = Path(dir_)
    paths = dir_.glob("**/*")
    return not any(True for path in paths if filter_(path))


def _ignore(path: Path) -> bool:
    path = path.resolve()
    if path.is_file() and path.name.startswith("."):
        return True
    if path.is_dir() and path.name in (
        ".jukit",
        ".ipynb_checkpoints",
        ".mypy_cache",
        ".pytest_cache",
        ".mtj.tmp",
        "__pycache__",
    ):
        return True
    return False


def remove_ess_empty(path: str | Path, ignore: Callable = _ignore) -> list[Path]:
    """Remove essentially empty directories under a path.

    :param path: The path to the directory to check.
    :param ignore: A bool function which returns True on files/directories to ignore.
    :return: A list of Path objects which failed to be removed.
    """
    fail = []
    for p in find_ess_empty(path, ignore=ignore):
        try:
            if p.is_file() or p.is_symlink():
                p.unlink()
            else:
                shutil.rmtree(p)
        except PermissionError:
            fail.append(p)
    return fail


def find_ess_empty(path: str | Path, ignore: Callable = _ignore) -> list[Path]:
    """Find essentially empty sub directories under a directory.

    :param path: The path to the directory to check.
    :param ignore: A bool function which returns True on files/directories to ignore.
    :return: A list of directories which are essentially empty.
    """
    if isinstance(path, str):
        path = Path(path)
    ess_empty = {}
    ess_empty_dir = []
    _find_ess_empty(
        path=path, ignore=ignore, ess_empty=ess_empty, ess_empty_dir=ess_empty_dir
    )
    return ess_empty_dir


def _find_ess_empty(
    path: Path, ignore: Callable, ess_empty: dict[Path, bool], ess_empty_dir: list[Path]
):
    if is_ess_empty(path=path, ignore=ignore, ess_empty=ess_empty):
        ess_empty_dir.append(path)
        return
    for p in path.iterdir():
        if p.is_dir():
            _find_ess_empty(
                path=p, ignore=ignore, ess_empty=ess_empty, ess_empty_dir=ess_empty_dir
            )


def is_ess_empty(
    path: Path, ignore: Callable = _ignore, ess_empty: dict[Path, bool] | None = None
):
    """Check if a directory is essentially empty.

    :param path: The path to the directory to check.
    :param ignore: A bool function which returns True on files/directories to ignore.
    :param ess_empty: A dictionary caching essentially empty paths.
    :raises FileNotFoundError: If the given path does not exist.
    :return: True if the directory is essentially empty and False otherwise.
    """
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"The file {path} does not exist!")
    if not os.access(path, os.R_OK):
        return False
    if path.is_symlink():
        return True
    path = path.resolve()
    if ess_empty is None:
        ess_empty = {}
    if path in ess_empty:
        return ess_empty[path]
    if ignore(path):
        return True
    for p in path.iterdir():
        if ignore(p):
            continue
        if p.is_file():
            return False
        if not is_ess_empty(p, ignore=ignore, ess_empty=ess_empty):
            ess_empty[path] = False
            return False
    ess_empty[path] = True
    return True


def append_lines(
    path: Path,
    lines: str | Iterable[str],
    exist_skip: bool = True,
) -> None:
    """Update a text file using regular expression substitution.

    :param path: A Path object to the file to be updated.
    :param lines: A (list of) line(s) to append.
        Note that "\\n" is automatically added to the end of each line to append.
    :param exist_skip: Skip if lines to append already exists in the file.
    """
    if isinstance(path, str):
        path = Path(path)
    text = path.read_text(encoding="utf-8")
    if not isinstance(lines, str):
        lines = "\n".join(lines)
    if not exist_skip or lines not in text:
        text += lines
    path.write_text(text, encoding="utf-8")


def replace_patterns(
    path: Path,
    pattern: str | Iterable[str],
    repl: str | Iterable[str],
    regex: bool = True,
) -> None:
    """Update a text file by replacing patterns with specified substitutions.

    :param path: A Path object to the file to be updated.
    :param pattern: A (list of) patterns to replace.
    :param repl: A (list of) replacements.
        or a function to map patterns to replacements.
    :param regex: If true, treat patterns as regular expression pattern;
        otherwise, perform exact matches.
    """
    if isinstance(path, str):
        path = Path(path)
    text = path.read_text(encoding="utf-8")
    if isinstance(pattern, str):
        pattern = [pattern]
    if isinstance(repl, str):
        repl = [repl]
    if regex:
        for p, r in zip(pattern, repl):
            text = re.sub(p, r, text)
    else:
        for p, r in zip(pattern, repl):
            text = text.replace(p, r)
    path.write_text(text, encoding="utf-8")


def get_files(dir_: str | Path, exts: str | list[str]) -> Iterable[Path]:
    """Get files with the specified file extensions.

    :param dir_: The path to a directory.
    :param exts: A (list of) file extensions (e.g., .txt).
    :yield: A generator of Path objects.
    """
    if isinstance(dir_, str):
        dir_ = Path(dir_)
    if isinstance(exts, str):
        exts = [exts]
    yield from _get_files(dir_, exts)


def _get_files(dir_: Path, exts: list[str]) -> Iterable[Path]:
    for path in dir_.iterdir():
        if path.is_file():
            if path.suffix.lower() in exts:
                yield path
        else:
            yield from _get_files(path, exts)


def has_header(
    files: str | Path | list[str | Path], num_files_checking: int = 5
) -> bool:
    """Check whether the files have headers.

    :param files: the list of files to check.
    :param num_files_checking: the number of non-empty files to use to decide whether there are header lines.
    :return: True if the files have headers and False otherwise.
    """
    # i: file index
    for i in range(len(files)):
        with open(files[i], "r", encoding="utf-8") as fin:
            first_line = fin.readline()
            if first_line:
                possible_header = first_line
                break
    # k: current number of non-empty files
    k = 1
    for j in range(i, len(files)):
        if k >= num_files_checking:
            break
        with open(files[j], "r", encoding="utf-8") as fin:
            first_line = fin.readline()
            if first_line:
                k += 1
                if first_line != possible_header:
                    return False
    return True


def _merge_with_headers(
    files: str | Path | list[str | Path], output: str | Path = ""
) -> None:
    """Merge files with headers. Keep only one header.

    :param files: A list of files
        or the path to a directory containing a list of files to merge.
    :param output: output files for merging the files.
    """
    with open(output, "wb") if output else sys.stdout.buffer as out:
        with open(files[0], "rb") as fin0:
            for line in fin0:
                out.write(line)
        for file in files[1:]:
            with open(file, "rb") as fin:
                fin.readline()
                for line in fin:
                    out.write(line)


def _merge_without_header(
    files: str | Path | list[str | Path], output: str | Path = ""
) -> None:
    """Merge files without header.

    :param files: A list of files
        or the path to a directory containing a list of files to merge.
    :param output: output files for merging the files.
    """
    with open(output, "wb") if output else sys.stdout.buffer as fout:
        for file in files:
            with open(file, "rb") as fin:
                for line in fin:
                    fout.write(line)
                fout.write(b"\n")


def merge(
    files: str | Path | list[str | Path],
    output: str = "",
    num_files_checking: int = 5,
) -> None:
    """Merge files. If there are headers in files, keep only one header in the single merged file.

    :param files: A list of files
        or the path to a directory containing a list of files to merge.
    :param output: output files for merging the files.
    :param num_files_checking: number of files for checking whether there are headers in files.
    """
    if isinstance(files, str):
        files = Path(files)
    if isinstance(files, Path):
        files = list(files.iterdir())
    if num_files_checking <= 0:
        num_files_checking = 5
    num_files_checking = min(num_files_checking, len(files))
    if has_header(files, num_files_checking):
        _merge_with_headers(files, output)
        return
    _merge_without_header(files, output)


def dedup_header(file: str | Path, output: str | Path = "") -> None:
    """Dedup headers in a file (due to the hadoop getmerge command).
    Only the header on the first line is kept and headers (identical line to the first line)
    on other lines are removed.

    :param file: The path to the file to be deduplicated.
    :param output: The path of the output file.
        If empty, then output to the standard output.
    """
    with (
        open(file, "rb") as fin,
        open(output, "wb") if output else sys.stdout.buffer as fout,
    ):
        header = fin.readline()
        fout.write(header)
        for line in fin:
            if line != header:
                fout.write(line)


def select(
    path: str | Path,
    columns: str | list[str],
    delimiter: str,
    output: str = "",
):
    """Select fields by name from a delimited file (not necessarily well structured).

    :param path: To path to a file (containing delimited values in each row).
    :param columns: A list of columns to extract from the file.
    :param delimiter: The delimiter of fields.
    :param output: The path of the output file.
        If empty, then output to the standard output.
    """
    if isinstance(path, str):
        path = Path(path)
    if isinstance(columns, str):
        columns = [columns]
    with path.open("r", encoding="utf-8") as fin:
        header = fin.readline().split(delimiter)
        index = []
        columns_full = []
        for idx, field in enumerate(header):
            if field in columns:
                index.append(idx)
                columns_full.append(field)
        with open(output, "w", encoding="utf-8") if output else sys.stdout as fout:
            fout.write(delimiter.join(columns_full) + "\n")
            for line in fin:
                fields = line.split(delimiter)
                fout.write(delimiter.join([fields[idx] for idx in index]) + "\n")


def prune_json(input: str | Path, output: str | Path = ""):
    """Prune fields (value_counts) from a JSON file.

    :param input: The path to a JSON file to be pruned.
    :param output: The path to output the pruned JSON file.
    """
    logger.info("Pruning the JSON fiel at {}...", input)
    if isinstance(input, str):
        input = Path(input)
    if isinstance(output, str):
        if output:
            output = Path(output)
        else:
            output = input.with_name(input.stem + "_prune.json")
    skip = False
    with (
        input.open("r", encoding="utf-8") as fin,
        output.open("w", encoding="utf-8") as fout,
    ):
        for line in fin:
            line = line.strip()
            if line == '"value_counts": {':
                skip = True
                continue
            if skip:
                if line in ("}", "},"):
                    skip = False
            else:
                fout.write(line)
    logger.info("The pruned JSON file is written to {}.", output)


def _filter_num(path: str | Path, pattern: str, num_lines: int):
    if isinstance(path, str):
        path = Path(path)
    results = []
    res = []
    count = 0
    for line in path.open(encoding="utf-8"):
        if count > 0:
            res.append(line)
            count -= 1
            continue
        if re.search(pattern, line):
            if res:
                results.append(res)
            res = []
            res.append(line)
            count = num_lines
    if res:
        results.append(res)
    return results


def _filter_sp(path: str | Path, pattern: str, sub_pattern: str):
    if isinstance(path, str):
        path = Path(path)
    results = []
    res = []
    sub = False
    for line in path.open(encoding="utf-8"):
        if sub:
            if re.search(sub_pattern, line):
                res.append(line)
            else:
                sub = False
        if re.search(pattern, line):
            if res:
                results.append(res)
            res = []
            res.append(line)
            sub = True
    if res:
        results.append(res)
    return results


def filter(
    path: str | Path, pattern: str, sub_pattern: str = "", num_lines: int = 0
) -> list[list[str]]:
    """Filter lines from a file.
    A main regex pattern is used to identify main rows.
    For each matched main row,
    a sub regex pattern or a fixed number of lines can be provided.
    If a sub regex pattern is provided,
    then lines matching the sub regex pattern following a main line are kept together with the main line.
    If a fixed number of lines is provided, e.g., ``num_lines=k``,
    then ``k`` additional lines after a main line are kept together with the main line.

    :param path: Path to a text file from which to filter lines.
    :param pattern: The main regex pattern.
    :param sub_pattern: The sub regex pattern (defaults to "", i.e., no sub pattern by default).
    :param num_lines: The num of additional lines (0 by default) to keep after a main line.
    :return: A list of list of lines.
    """
    if sub_pattern:
        return _filter_sp(path, pattern=pattern, sub_pattern=sub_pattern)
    return _filter_num(path, pattern=pattern, num_lines=num_lines)


def trace_dir_upwards(path: str | Path, name: str) -> PosixPathPair:
    """Find the parent directory with the specified name.

    Args:
        path: A local path contains `/name/`.
        name: The base name (stem) of the parent directory.

    Returns:
        A PosixPathPair which contains the parent directory
        and the relative path to this parent directory.
    """

    def _trace_dir_upwards(path: Path) -> Path:
        while (stem := path.stem) != name:
            if not stem:
                raise ValueError(f"The path {path} does not contain /{name}/!")
            path = path.parent
        return path

    if isinstance(path, str):
        path = Path(path)
    prefix = _trace_dir_upwards(path)
    return PosixPathPair(prefix, path.relative_to(prefix))



"""Module for log filtering."""

from typing import Sequence, TextIO
from pathlib import Path
import sys
import re
from collections import deque
from difflib import SequenceMatcher
from tqdm import tqdm
from loguru import logger

DASH_50 = "-" * 50
DASH_100 = "-" * 100


class LogDeduper:
    """Dedup similar log lines."""

    def __init__(self, threshold: float = 0.7):
        self.lines = []
        self._threshold = threshold

    def similarity(self, line):
        """Calcualte similarity between 2 lines.

        :param line1: A line of logging message.
        :param line2: Another line of logging message.
        :return: A similarity score (between 0 and 1) between the 2 lines.
        """
        return max(
            (
                SequenceMatcher(None, line, target).ratio()
                for target in reversed(self.lines)
            ),
            default=0,
        )

    def add(self, line, line_num):
        """Add a line.

        :param line: A line of logging message.
        :param line_num: The row number (0-based) of the line.
        """
        if self.similarity(line) < self._threshold:
            self.lines.append(f"L{line_num}: {line}")

    def write(self, fout: TextIO):
        """Write deduplicated log into a file.

        :param fout: A file handler for outputing log.
        """
        for line in self.lines:
            fout.write(line)


class LogFilter:
    """A class for log filtering."""

    KEYWORDS = (
        "spark.yarn.executor.memoryOverhead",
        "not found",
        "OOM",
        "Error",
        "error",
        "Exception",
        "exception",
        "failures",
    )
    PATTERNS = (
        (r"\d{2,}[-/]\d{2,}[-/]\d{2,}\s\d+:\d+:\d+", "YYYY/MM/DD HH:MM:SS"),
        (r"\d{,3}\.\d{,3}\.\d{,3}\.\d{,3}(:\d+)?", "0.0.0.0:0"),
        (r"streamId=\d+", "streamId=*"),
        (r"chunkIndex=\d+", "chunkIndex=*"),
        (r"send RPC \d+", "send RPC *"),
    )

    def __init__(
        self,
        log_file: str | Path,
        context_size: int = 5,
        keywords: Sequence[str] = KEYWORDS,
        patterns: Sequence[tuple[str, str]] = PATTERNS,
        output: str | Path = "",
        threshold: float = 0.7,
        dump_by_keyword: bool = False,
    ):
        self._log_file = (
            log_file if isinstance(log_file, Path) else Path(log_file)
        ).resolve()
        self._context_size: int = context_size
        self._keywords: Sequence[str] = keywords
        self._patterns: Sequence[tuple[str, str]] = patterns
        self._num_rows: int = 0
        self._lookup: dict[str, dict[str, int]] = {kwd: {} for kwd in self._keywords}
        self._queue: deque = deque()
        self._output: Path = self._get_output(output)
        self._threshold: float = threshold
        self._dump_by_keyword: bool = dump_by_keyword

    def _get_output(self, output: str | Path) -> Path:
        """Get a valid output file.

        :param output: The path to the output file.
        """
        if output == "" or Path(output).resolve() == self._lookup:
            return self._log_file.with_name(
                self._log_file.stem + "_s" + self._log_file.suffix
            )
        if isinstance(output, str):
            output = Path(output)
        return output.resolve()

    def _regularize(self, line) -> str:
        """Get rid of substrings with patterns specified by the regular expressions.

        :param line: A line of logging message.
        :return: The regularized the line message.
        """
        for pattern, replace in self._patterns:
            line = re.sub(pattern, replace, line)
        return line

    def _dump_queue(self, lines) -> None:
        """Dump content in the queue.

        :param lines: A list to dump the queue to.
        """
        lines.append(DASH_100 + "\n")
        lines.extend(self._queue)
        self._queue.clear()

    def _keep(self, idx: int, line: str) -> bool:
        """Check whether the line should be kept.

        :param idx: The original row number (0-based) of the line.
        :param line: A line of logging message.
        :return: True if the line is to be kept and False otherwise.
        """
        if " ./" in line or "-XX:OnOutOfMemoryError=" in line:
            return False
        for kwd in self._keywords:
            if kwd in line:
                line = self._regularize(line)
                if line in self._lookup[kwd]:
                    return False
                self._lookup[kwd][line] = idx
                return True
        return False

    def _count_rows(self):
        """Count the total number of rows."""
        if self._num_rows:
            return
        logger.info("Counting total number of rows ...")
        with open(self._log_file, "r", encoding="utf-8") as fin:
            self._num_rows = sum(1 for line in fin)
        logger.info("Total number of rows: {:,}", self._num_rows)

    def _scan_error_lines(self) -> None:
        print()
        logger.info("Scanning for error lines in the log ...")
        lines = [DASH_50 + " Possible Error Lines " + DASH_50 + "\n"]
        with open(self._log_file, "r", encoding="utf-8") as fin:
            dump_flag = -1
            for idx, line in tqdm(enumerate(fin), total=self._num_rows):
                self._queue.append(f"L{idx}: {line}")
                keep = self._keep(idx, line)
                if keep:
                    dump_flag = 0
                    continue
                if dump_flag == -1:
                    if len(self._queue) > self._context_size:
                        self._queue.popleft()
                    continue
                dump_flag += 1
                if dump_flag >= self._context_size:
                    self._dump_queue(lines)
                    dump_flag = -1
            if dump_flag >= 0:
                self._dump_queue(lines)
        with open(self._output, "w", encoding="utf-8") as fout:
            fout.writelines(lines)
        logger.info("Possible Error Lines have been dumped into {}", self._output)

    def filter(self) -> None:
        """Filter informative liens from a Spark application log."""
        self._count_rows()
        self._scan_error_lines()
        self._dedup_log()

    def _dedup_log(self):
        print()
        # create dir for dumping errors by keyword
        if self._dump_by_keyword:
            dir_ = self._output.parent / (self._log_file.stem + "_k")
            dir_.mkdir(parents=True, exist_ok=True)
            logger.info(
                "Error lines will be dumped by keyword into the directory {}.", dir_
            )
        else:
            dir_ = None
        # dedup error lines
        lines_unique = []
        for kwd, lines in self._lookup.items():
            if not lines:
                continue
            logger.info('Deduplicating error lines corresponding to "{}" ...', kwd)
            lines_unique.extend(self._dedup_log_1(kwd, lines, dir_))
        lines_unique = [self._error_priority(line) for line in lines_unique]
        lines_unique.sort()
        with self._output.open("a", encoding="utf-8") as fout:
            self._write_lines_unique(lines_unique, fout)
        self._write_lines_unique(lines_unique, sys.stdout)

    @staticmethod
    def _write_lines_unique(lines_unique: list[tuple[int, str, str]], fout: TextIO):
        fout.write("\n" + DASH_50 + " Deduped Error Lines " + DASH_50 + "\n")
        fout.write(
            "https://www.legendu.net/misc/blog/A-comprehensive-list-of-issues-in-spark-applications\n\n"
        )
        for _, line, url in lines_unique:
            fout.write(line)
            fout.write(f"Possible causes and solutions: {url}\n\n")

    @staticmethod
    def _error_priority(line: str) -> tuple[int, str, str]:
        """Return priority (with a smaller value means higher priority) of an error line.

        :param line: An error line.
        :return: The priority of the error line.
        """
        patterns = [
            ("SIGILL", 1, "https://www.legendu.net/en/blog/spark-issue:-shell-related"),
            (
                "(?i)command not found",
                1,
                "https://www.legendu.net/en/blog/spark-issue:-shell-related",
            ),
            (
                "(?i)panicked at.*RUST_BACKTRACE",
                1,
                "https://www.legendu.net/en/blog/spark-issue:-rust-panic/",
            ),
            (
                "(?i)libc.*not found",
                1,
                "https://www.legendu.net/misc/blog/spark-issue-libc-not-found/",
            ),
            (
                r"(?i)ArrowInvalid",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                r"(?i)ArrowTypeError: Expected",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-ArrowTypeError:-Expect-a-type-but-got-a-different-type",
            ),
            (
                r"(?i)Arrow legacy IPC format is not supported",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-RuntimeError:-Arrow-legacy-IPC-format-is-not-supported",
            ),
            (
                r"(?i)TypeError: .*has no len()",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                r"(?i)ValueError: min() arg is an empty sequence",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                r"(?i)CalledProcessError: Command .* returned non-zero exit status",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                r"(?i)error: Found argument .* which wasn't expected",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                r"(?i)RuntimeError: Result vector from pandas_udf was not the required length",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)ViewFs: Cannot initialize: Empty Mount table in config",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-ViewFs:-Cannot-initialize:-Empty-Mount-table-in-config",
            ),
            (
                "(?i)IllegalArgumentException: Wrong FS",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-IllegalArgumentException:-Wrong-FS",
            ),
            (
                "(?i)object has no attribute",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)No such file or directory:",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)error: the following arguments are required:",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)error: unrecognized arguments:",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)error: argument",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)ModuleNotFoundError: No module named",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)SyntaxError: invalid syntax",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)NameError: name .* is not defined",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (
                "(?i)IndentationError: unexpected indent",
                1,
                "https://www.legendu.net/misc/blog/Spark-issue:-Pure-Python-code-errors",
            ),
            (r"SIGBUS", 1, "https://www.legendu.net/misc/blog/spark-issue:-SIGBUS"),
            (
                "(?i)NSQuotaExceededException",
                1,
                "https://www.legendu.net/en/blog/spark-issue:-namespace-quota-is-exceeded",
            ),
            (
                r"(?i)RuntimeException: Unsupported literal type class",
                1,
                " https://www.legendu.net/en/blog/spark-issue:-RuntimeException:-unsupported-literal-type-class",
            ),
            (
                r"(?i)TypeError: withReplacement",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-TypeError-withReplacement",
            ),
            (
                "(?i)URISyntaxException",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-URISyntaxException",
            ),
            (
                "(?i)Could not find any configured addresses for URI",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-RuntimeException:-Could-not-find-any-configured-addresses-for-URI",
            ),
            (
                "(?i)table not found",
                1,
                "https://www.legendu.net/misc/blog/spark-issue-table-not-found/",
            ),
            (
                "(?i)SparkContext: A master URL must be set",
                1,
                "https://www.legendu.net/misc/blog/spark-issue-a-master-url-must-be-set-in-your-configuration/",
            ),
            (
                r"(?i)org\.apache\.spark\.sql\.AnalysisException.*cannot resolve",
                1,
                "https://www.legendu.net/misc/blog/spark-issue-AnalysisException-cannot-resolve",
            ),
            (
                r"(?i)org\.apache\.spark\.sql\.AnalysisException.*Path does not exist",
                1,
                "https://www.legendu.net/misc/blog/spark-issue-AnalysisException-Path-does-not-exist",
            ),
            (
                r"(?i)org\.apache\.hadoop\.security\.AccessControlException.*Permission denied",
                1,
                "https://www.legendu.net/misc/blog/Spark-Issue-AccessControlException-Permission-denied",
            ),
            (
                r"(?i)org\.apache\.spark\.InsertOperationConflictException.*Failed to hold insert operation lock",
                1,
                "https://www.legendu.net/misc/blog/spark-issue-InsertOperationConflictException-failed-to-hold-insert-operation-lock",
            ),
            (
                r"(?i)max number of executor failures",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-max-number-of-executor-failures-reached",
            ),
            (
                r"(?i)IllegalArgumentException: System memory \d* must be at least",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-IllegalArgumentException:-System-memory-must-be-at-least",
            ),
            (
                r"(?i)InvalidResourceRequestException",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-InvalidResourceRequestException",
            ),
            (
                r"(?i)The quota system is disabled",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-getQuotaUsage",
            ),
            (
                r"(?i)AnalysisException: Found duplicate column(s)",
                1,
                "https://www.legendu.net/misc/blog/spark-issue:-AnalysisException:-Found-duplicated-columns",
            ),
            (
                "(?i)broadcastTimeout",
                2,
                "https://www.legendu.net/misc/blog/spark-issue:-could-not-execute-broadcast-in-300s",
            ),
            (
                "(?i)serialized results is bigger than spark.driver.maxResultSize",
                2,
                "https://www.legendu.net/misc/blog/spark-issues-total-size-bigger-than-maxresultsize/",
            ),
            (
                "(?i)Block rdd_.* could not be removed as it was not found on disk or in memory",
                2,
                "https://www.legendu.net/misc/blog/spark-issue:-block-could-not-be-removed-as-it-was-not-found-on-disk-or-in-memory",
            ),
            (
                r"(?i)java\.io\.FileNotFoundException",
                2,
                "https://www.legendu.net/misc/blog/spark-issue-file-not-found-exception/",
            ),
            (
                "(?i)Container killed by YARN for exceeding memory limits",
                2,
                "https://www.legendu.net/misc/blog/spark-issue-Container-killed-by-YARN-for-exceeding-memory-limits/",
            ),
            (
                r"(?i)org\.apache\.spark\.memory\.SparkOutOfMemoryError",
                3,
                "https://www.legendu.net/misc/tag/spark-issue.html",
            ),
            (
                "(?i)Container from a bad node",
                3,
                "https://www.legendu.net/misc/tag/spark-issue.html",
            ),
            (
                "(?i)No live nodes contain block BP",
                3,
                "https://www.legendu.net/misc/tag/spark-issue.html",
            ),
            (
                "(?i)ReplicaNotFoundException: Replica not found for",
                3,
                "https://www.legendu.net/misc/tag/spark-issue.html",
            ),
        ]
        for pattern, priority, url in patterns:
            if re.search(pattern, line):
                return priority, line, url
        return 10, line, "https://www.legendu.net/misc/tag/spark-issue.html"

    def _dedup_log_1(
        self, kwd: str, lines: dict[str, int], dir_: Path | None
    ) -> list[str]:
        deduper = LogDeduper(self._threshold)
        lines = sorted(lines.items())
        for line, idx in tqdm(lines):
            deduper.add(line, idx)
        # deduper.write(sys.stdout)
        # deduper.write(fout)
        if self._dump_by_keyword:
            with (dir_ / kwd).open("w") as fout_kwd:
                for line, idx in lines:
                    fout_kwd.write(f"L{idx}: {line}\n")
        return deduper.lines



#!/usr/bin/env python3
"""Make it easier to authenticate users' personal accounts on Hadoop.
If an user specifiy a password when authenticating,
the password is encrypted and saved into a profile that is readable/writable only by the user.
If an user authenticate without specifying password,
the saved password is used so that users do not have to type in password to authenticate every time.
"""

from typing import Any
import os
from pathlib import Path
import socket
import datetime
import getpass
import time
import subprocess as sp
import base64
from argparse import ArgumentParser, Namespace
import yaml
from loguru import logger
import notifiers

HOME = Path.home()
PROFILE = HOME / ".pykinit_profile"
HOST = socket.gethostname()
HOST_IP = socket.gethostbyname(HOST)
USER = getpass.getuser()
PID = os.getpid()


def save_passwd(passwd: str) -> None:
    """Encrypt and save the password into a profile that is readable/writable only by the user.

    :param passwd: The password of the user (to save).
    """
    bytes_ = passwd.encode("ascii")
    encode = base64.b64encode(bytes_).decode()
    with open(PROFILE, "w", encoding="utf-8") as fout:
        fout.write(encode)
    os.chmod(PROFILE, 0o600)


def read_passwd() -> str:
    """Read in the saved password.

    :return: Password in the profile file or empty string if the file does not exist.
    """
    if os.path.isfile(PROFILE):
        os.chmod(PROFILE, 0o600)
        with open(PROFILE, "r", encoding="utf-8") as fin:
            return base64.b64decode(fin.read()).decode()
    passwd = getpass.getpass("Please enter password for kinit: ")
    save_passwd(passwd)
    return passwd


def _warn_passwd_expiration(process, email: dict[str, str]):
    lines = process.stdout.decode().split("\n")
    msg = "\n".join(
        line
        for line in lines
        if line.strip().startswith("Warning: Your password will expire")
    )
    if msg and email:
        subject = "Your Hadoop cluster password is expiring!"
        notifiers.get_notifier("email").notify(
            from_=email["from"],
            to=email["to"],
            subject=subject,
            message=msg,
            host=email["host"],
            username="",
            password="",
        )


def authenticate(password: str, email: dict[str, str], user: str = "") -> None:
    """Authenticate using the shell command /usr/bin/kinit.

    :param password: The password of the user.
    :param email: A dict containing email information ("from", "to" and "host").
    :param user: User name. If empty, the current user name is used.
    """
    SUBJECT = "kinit: authentication {}"
    MSG = f"kinit ({PID}): authentication on {HOST} ({HOST_IP}) {'{}'} at {datetime.datetime.now()}"
    try:
        process = sp.run(
            ["/usr/bin/kinit", user if user else USER],
            input=password.encode(),
            check=True,
            capture_output=True,
        )
        subject = SUBJECT.format("succeeded")
        msg = MSG.format("succeeded")
        logger.info(msg)
        if email:
            notifiers.get_notifier("email").notify(
                from_=email["from"],
                to=email["to"],
                subject=subject,
                message=msg,
                host=email["host"],
                username="",
                password="",
            )
        _warn_passwd_expiration(process, email)
    except sp.CalledProcessError:
        subject = SUBJECT.format("failed")
        msg = MSG.format("failed")
        logger.warning(msg)
        if email:
            notifiers.get_notifier("email").notify(
                from_=email["from"],
                to=email["to"],
                subject=subject,
                message=msg,
                host=email["host"],
                username="",
                password="",
            )


def parse_args(args=None, namespace=None) -> Namespace:
    """Parse command-line arguments for the script.

    :param args: The arguments to parse.
    If None, the command-line arguments are parsed.
    :param namespace: An initial Namespace object.
    :return: A Namespace object containing parsed command-line options.
    """
    parser = ArgumentParser(description="Easy kinit authentication.")
    parser.add_argument(
        "-u",
        "--user",
        dest="user",
        default="",
        help="The name of the user to authenticate.",
    )
    parser.add_argument(
        "-p", "--password", dest="password", default="", help="the user's password."
    )
    parser.add_argument(
        "-m",
        "--minute",
        dest="minute",
        type=int,
        default=None,
        help="Run the script as a deamon with the specified frequency (in minutes).",
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        type=Path,
        default=HOME / ".pykinit.yaml",
        help="The path to a configure file which contains email information.",
    )
    return parser.parse_args(args, namespace)


def _read_config(config: Path | str) -> dict[str, Any]:
    if isinstance(config, str):
        config = Path(config)
    if not config.is_file():
        return {"email": {}}
    with config.open("r", encoding="utf-8") as fin:
        return yaml.load(fin, Loader=yaml.FullLoader)


def main() -> None:
    """Authenticate the user using either supplied or saved password.

    :raises ExceptionNoPassword: If no password is provided or found.
    """
    args = parse_args()
    if args.password:
        save_passwd(args.password)
    password = read_passwd()
    config = _read_config(args.config)
    if args.minute:
        while True:
            authenticate(read_passwd(), config["email"])
            time.sleep(args.minute * 60)
    else:
        authenticate(password, config["email"], args.user)


if __name__ == "__main__":
    main()



"""Wrapping HDFS commands."""

from pathlib import Path
import subprocess as sp
import pandas as pd
from loguru import logger
from ..shell import to_frame
from ..filesystem import count_path


class Hdfs:
    """A class abstring the hdfs command."""

    def __init__(self, bin: str = "/apache/hadoop/bin/hdfs"):
        self._bin = bin

    def ls(
        self, path: str, dir_only: bool = False, recursive: bool = False
    ) -> pd.DataFrame:
        """Return the results of hdfs dfs -ls /hdfs/path as a DataFrame.

        :param path: A HDFS path.
        :param dir_only: If true, list directories only.
        :param recursive: If True, list content of the HDFS path recursively.
        :return: The result of hdfs dfs -ls as a pandas DataFrame.
        """
        cols = [
            "permissions",
            "replicas",
            "userid",
            "groupid",
            "bytes",
            "mdate",
            "mtime",
            "path",
        ]
        cmd = f"{self._bin} dfs -ls "
        if dir_only:
            cmd += "-d "
        if recursive:
            cmd += "-R "
        cmd += path
        logger.info("Running command: {}. Might take a while.", cmd)
        frame = to_frame(cmd, split=r" +", skip=() if dir_only else [0], header=cols)
        frame.bytes = frame.bytes.astype(int)
        frame.mtime = pd.to_datetime(frame.mdate + " " + frame.mtime)
        frame.drop("mdate", axis=1, inplace=True)
        return frame

    def count(self, path: str) -> pd.DataFrame:
        """Return the results of hdfs dfs -count -q -v /hdfs/path as a DataFrame.

        :param path: A HDFS path.
        :return: The result of hdfs dfs -count as a pandas DataFrame.
        """
        cmd = f"{self._bin} dfs -count -q -v {path}"
        logger.info("Running command: {}. Might take a while.", cmd)
        frame = to_frame(cmd, split=r" +", header=0)
        frame.columns = frame.columns.str.lower()
        return frame

    def du(self, path: str, depth: int = 1) -> pd.DataFrame:
        """Get the size of HDFS paths.

        :param path: A HDFS path.
        :param depth: The depth (by default 1) of paths to calculate sizes for.
            Note that any depth less than 1 is treated as 1.
        :return: Disk usage of the HDFS path as a pandas DataFrame.
        """
        index = len(path.rstrip("/"))
        if depth > 1:
            paths = self.ls(path, recursive=True).path
            frames = [
                self._du_helper(path)
                for path in paths
                if path[index:].count("/") + 1 == depth
            ]
            return pd.concat(frames, ignore_index=True)
        return self._du_helper(path)

    def _du_helper(self, path: str) -> pd.DataFrame:
        cmd = f"{self._bin} dfs -du {path}"
        logger.info("Running command: {}. Might take a while.", cmd)
        frame = to_frame(cmd, split=r" +", header=["bytes", "path"])
        frame.bytes = frame.bytes.astype(int)
        return frame

    def exists(self, path: str) -> bool:
        """Check if a HDFS path exist.

        :param path: A HDFS path.
        :return: True if the HDFS path exists and False otherwise.
        """
        cmd = f"{self._bin} dfs -test -e {path}"
        logger.info("Running command: {}. Might take a while.", cmd)
        try:
            sp.run(cmd, shell=True, check=True)
            return True
        except sp.CalledProcessError:
            return False

    def remove(self, path: str) -> None:
        """Remove a HDFS path.
        :param path: A HDFS path.
        """
        cmd = f"{self._bin} dfs -rm -r {path}"
        logger.info("Running command: {}. Might take a while.", cmd)
        sp.run(cmd, shell=True, check=True)

    def num_partitions(self, path: str) -> int:
        """Get the number of partitions of a HDFS path.

        :param path: A HDFS path.
        :return: The number of partitions under the HDFS path.
        """
        cmd = f"{self._bin} dfs -ls {path}/part-* | wc -l"
        logger.info("Running command: {}. Might take a while.", cmd)
        return int(sp.check_output(cmd, shell=True))

    def get(
        self, hdfs_path: str, local_dir: str | Path = "", is_file: bool = False
    ) -> None:
        """Download data from HDFS into a local directory.

        :param hdfs_path: The HDFS path (can be both a file or a directory) to copy.
        :param local_dir: The local directory to copy HDFS files into.
        :param is_file: A boolean indicator of whether the HDFS path is a file or a directory.
        """
        if isinstance(local_dir, str):
            local_dir = Path(local_dir)
        local_dir.mkdir(parents=True, exist_ok=True)
        if is_file:
            cmd = f"{self._bin} dfs -get {hdfs_path} {local_dir}"
        else:
            cmd = f"{self._bin} dfs -get {hdfs_path}/* {local_dir}"
        logger.info("Running command: {}. Might take a while.", cmd)
        sp.run(cmd, shell=True, check=True)
        print(
            f"Content of the HDFS path {hdfs_path} has been fetch into the local directory {local_dir}"
        )

    def count_path(self, path: str) -> pd.Series:
        """Count frequence of paths and their parent paths.

        :param path: An iterable collection of paths.
        :return: Frequency of paths as a pandas Series.
        """
        frame = self.ls(path, recursive=True)
        return count_path(frame.path)

    def mkdir(self, path: str) -> None:
        """Create a HDFS path.

        :param path: The HDFS path to create.
        """
        cmd = f"{self._bin} dfs -mkdir -p {path}"
        logger.info("Running command: {}. Might take a while.", cmd)
        sp.run(cmd, shell=True, check=True)
        logger.info(f"The HDFS path {path} has been created.")

    def put(
        self,
        local_path: str | Path,
        hdfs_path: str,
        create_hdfs_path: bool = False,
    ) -> None:
        """Copy data from local to HDFS.
        :param local_path: A local path to copy to HDFS.
        :param hdfs_path: The HDFS path/directory to copy data into.
        :param create_hdfs_path: If true, create the HDFS path if not exists.
        """
        if create_hdfs_path:
            self.mkdir(hdfs_path)
        cmd = f"{self._bin} dfs -put -f {local_path} {hdfs_path}"
        logger.info("Running command: {}. Might take a while.", cmd)
        sp.run(cmd, shell=True, check=True)
        logger.info(
            f"The local path {local_path} has been uploaded into the HDFS path {hdfs_path}"
        )

    def fetch_partition_names(
        self, path: str, extension: str = ".parquet"
    ) -> list[str]:
        """Get Parquet partition names (with the parent directory) under a HDFS path.

        :param path: A HDFS path.
        :param extension: Return partitions with the specified file extension.
        :return: A list of the partition names (with the parent directory).
        """
        paths = self.ls(path).path
        return [path for path in paths if path.endswith(extension)]

    def rm(self, path: str, recursive: bool = True, skip_trash: bool = False) -> bool:
        """Remove a HDFS path.

        :param path: A HDFS path to remove.
        :param recursive: If true, use the option -r.
        :param skip_trash: If true, use the option -skipTrash.
        :return: True if the path is removed successfully and false otherwise.
        """
        flag_skip_trash = "-skipTrash" if skip_trash else ""
        flag_recursive = "-r" if recursive else ""
        cmd = f"{self._bin} dfs -rm {flag_recursive} {flag_skip_trash} {path}"
        logger.info("Running command: {}. Might take a while.", cmd)
        proc = sp.run(cmd, shell=True)  # pylint: disable=W1510
        return proc.returncode == 0

    def rm_robust(
        self, path: str, skip_trash: bool = False, user: str = ""
    ) -> list[str]:
        """Remove a HDFS path in a robust way.
            If removing a directory fails,
            the method continues to remove its subfiles and subdirs
            insteadd of throwing an exception.

        :param path: A HDFS path to remove.
        :param skip_trash: If true, use the option -skipTrash.
        :param user: If specified, remove only paths belong to the user.
        :return: A list of HDFS paths that are successfully removed.
        """
        if user:
            frame = self.ls(path, dir_only=True)
            frame = frame[frame.userid == user]
            if frame.empty:
                return []
        if self.rm(path, recursive=True, skip_trash=skip_trash):
            return [path]
        frame = self.ls(path)
        if user:
            frame = frame[frame.userid == user]
        paths_removed = []
        # remove files
        for file in frame[~frame.permissions.str.startswith("d")].path:
            if self.rm(file, skip_trash=skip_trash):
                paths_removed.append(file)
        for p in frame[frame.permissions.str.startswith("d")].path:
            paths_removed.extend(self.rm_robust(p, skip_trash=skip_trash, user=user))
        return paths_removed



"""A module providing wrap over the git command."""

from pathlib import Path
import subprocess as sp


class Repo:
    """A local Git repository."""

    def __init__(self, root: str | Path):
        self._root = str(root)
        self._remote = self.remote()

    def remote(self) -> str:
        """Run the "git remote" command."""
        proc = sp.run(
            ["git", "-C", self._root, "remote"], check=True, capture_output=True
        )
        return proc.stdout.strip().decode()

    def pull(self, branch: str) -> None:
        """Run the "git pull" command."""
        sp.run(["git", "-C", self._root, "pull", self._remote, branch], check=True)

    def push(self, branch: str = "") -> None:
        """Run the "git push" command."""
        if not branch:
            branch = self.active_branch()
        sp.run(["git", "-C", self._root, "push", self._remote, branch], check=True)

    def _branch(self) -> list[str]:
        proc = sp.run(
            ["git", "-C", self._root, "branch"], check=True, capture_output=True
        )
        return proc.stdout.strip().decode().split("\n")

    def active_branch(self) -> str:
        """Get the active branch of the repository."""
        lines = self._branch()
        return next(line[2:] for line in lines if line.startswith("* "))

    def branch(self):
        """Run the "git branch" command to get all branches of the repository."""
        return [b[2:] for b in self._branch()]

    def add(self, pattern: str = "."):
        """Run the "git add" command."""
        sp.run(["git", "-C", self._root, "add", pattern], check=True)

    def commit(self, message: str):
        """Run the "git commit" command."""
        sp.run(["git", "-C", self._root, "commit", "-m", message], check=True)

    def status(self):
        """Run the "git status" command."""
        sp.run(["git", "-C", self._root, "status"], check=True)

    def tag(self, tag="") -> sp.CompletedProcess | list[str]:
        """Run the "git tag" command."""
        if tag:
            return sp.run(["git", "tag", tag], check=True)
        proc = sp.run(["git", "tag"], check=True, capture_output=True)
        return proc.stdout.strip().decode().split("\n")

    def checkout(self, branch: str = "", new: str = ""):
        """Run the "git checkout" command."""
        if new:
            return sp.run(["git", "checkout", "-b", new], check=True)
        if not branch:
            raise ValueError("Either new or branch is required!")
        return sp.run(["git", "checkout", branch], check=True)



#!/usr/bin/env python3
"""Jupyter/Lab notebooks related utils."""

from pathlib import Path
import subprocess as sp
import itertools as it
import nbformat
from nbconvert import HTMLExporter

HOME = Path.home()


def nbconvert_notebooks(root_dir: str | Path, cache: bool = False) -> None:
    """Convert all notebooks under a directory and its subdirectories using nbconvert.

    :param root_dir: The directory containing notebooks to convert.
    :param cache: If True, previously generated HTML files will be used if they are still update to date.
    """
    if isinstance(root_dir, str):
        root_dir = Path(root_dir)
    notebooks = root_dir.glob("**/*.ipynb")
    exporter = HTMLExporter()
    for notebook in notebooks:
        html = notebook.with_suffix(".html")
        if (
            cache
            and html.is_file()
            and html.stat().st_mtime >= notebook.stat().st_mtime
        ):
            continue
        code, _ = exporter.from_notebook_node(nbformat.read(notebook, as_version=4))
        html.write_text(code, encoding="utf-8")


def _get_jupyter_paths():
    proc = sp.run("jupyter --path", shell=True, check=True, capture_output=True)
    lines = proc.stdout.decode().strip().split("\n")
    lines = (line.strip() for line in lines)
    return [line for line in lines if line.startswith("/")]


def _find_path_content(path, pattern):
    if isinstance(path, str):
        path = Path(path)
    for p in path.glob("**/*"):
        if p.is_file():
            try:
                if pattern in p.read_text():
                    yield p
            except Exception:
                pass


def _find_path_path(path, pattern):
    if isinstance(path, str):
        path = Path(path)
    for p in path.glob("**/*"):
        if pattern in str(p):
            yield p


def find_jupyter_path(pattern, content: bool):
    """Find Jupyter/Lab paths match a pattern.

    :param pattern: The pattern to search for.
    :param content: If True, search file content for the pattern;
    otherwise, ssearch path name for the pattern.
    """
    paths = _get_jupyter_paths()
    if content:
        paths = [_find_path_content(path, pattern) for path in paths]
    else:
        paths = [_find_path_path(path, pattern) for path in paths]
    return list(it.chain.from_iterable(paths))



"""Date/time related utils."""

from typing import Iterable
import datetime
import pandas as pd

# import pysnooper
DATE_FORMAT_DASH = "%Y-%m-%d"
DATE_FORMAT_DIGIT = "%Y%m%d"
TIME_FORMAT_DASH = "%Y-%m-%d %H:%M:%S"
_MAPPING = {
    "Monday": 0,
    "Mon": 0,
    "Tuesday": 1,
    "Tue": 1,
    "Wednesday": 2,
    "Wed": 2,
    "Thursday": 3,
    "Thu": 3,
    "Friday": 4,
    "Fri": 4,
    "Saturday": 5,
    "Sat": 5,
    "Sunday": 6,
    "Sun": 6,
}


def range_str(
    start, stop, *, step=datetime.timedelta(days=1), fmt: str = TIME_FORMAT_DASH
) -> Iterable[str]:
    """Generate datetime range as str.

    :param start: A datetime object or a string that can be parsed into a datetime.
    :param stop: A datetime object or a string that can be parsed into a datetime.
    :param step: A timedelta object specifying how much the values in the sequence increase at each step.
    :param fmt: The format of date/time (defaults to TIME_FORMAT_DASH)
    :yield: A generator of datetime in string format.
    """
    for ts in range(start=start, stop=stop, step=step):
        yield ts.strftime(fmt)


def range(start, stop, step=datetime.timedelta(days=1)) -> Iterable[datetime.datetime]:
    """Generate a range of datetime objects.

    :param start: A datetime object or a string that can be parsed into a datetime.
    :param stop: A datetime object or a string that can be parsed into a datetime.
    :param step: A timedelta object specifying how much the values in the sequence increase at each step.
    :yield: A generator of datetime objects.
    """
    start = pd.to_datetime(start)
    stop = pd.to_datetime(stop)
    curr_dt = start
    while curr_dt < stop:
        yield curr_dt
        curr_dt += step


def _format_weekday(weekday: int | str) -> int:
    if isinstance(weekday, str):
        weekday = _MAPPING[weekday]
    return weekday


def last_weekday(
    date: datetime.date, weekday: int | str, exclude_date: bool = False
) -> datetime.date:
    """Get the date of the last occurrence of the specified weekday.

    :param date: The date from when to date back.
    :param weekday: An integer (0 stands for Monday)
        or the name (full or 3-letter abbreviation) of a weekday/weekend.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of specified last weekday.
    """
    diff = date.weekday() - _format_weekday(weekday)
    if diff < 0:
        diff += 7
    if exclude_date and diff == 0:
        diff = 7
    return date - datetime.timedelta(days=diff)


def last_monday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the last occurrence of Monday.

    :param date: The date from when to date back.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of last Monday.
    """
    return last_weekday(date=date, weekday="Mon", exclude_date=exclude_date)


def last_tuesday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the last occurrence of Tuesday.

    :param date: The date from when to date back.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of last Tuesday.
    """
    return last_weekday(date=date, weekday="Tue", exclude_date=exclude_date)


def last_wednesday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the last occurrence of Wednesday.

    :param date: The date from when to date back.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of last Wednesday.
    """
    return last_weekday(date=date, weekday="Wed", exclude_date=exclude_date)


def last_thursday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the last occurrence of Thursday.

    :param date: The date from when to date back.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of last Thursday.
    """
    return last_weekday(date=date, weekday="Thu", exclude_date=exclude_date)


def last_friday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the last occurrence of Friday.

    :param date: The date from when to date back.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of last Friday.
    """
    return last_weekday(date=date, weekday="Fri", exclude_date=exclude_date)


def last_saturday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the last occurrence of Saturday.

    :param date: The date from when to date back.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of last Saturday.
    """
    return last_weekday(date=date, weekday="Sat", exclude_date=exclude_date)


def last_sunday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the last occurrence of Sunday.

    :param date: The date from when to date back.
    :param exclude_date: Whether to exclude `date` when dating back.
    :return: The date of last Sunday.
    """
    return last_weekday(date=date, weekday="Sun", exclude_date=exclude_date)


def next_weekday(
    date: datetime.date, weekday: int | str, exclude_date: bool = False
) -> datetime.date:
    """Get the date of the next specified weekday.

    :param date: The date from when to date forward.
    :param weekday: An integer (0 stands for Monday)
        or the name (full or 3-letter abbreviation) of a weekday/weekend.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next weekday.
    """
    diff = _format_weekday(weekday) - date.weekday()
    if diff < 0:
        diff += 7
    if exclude_date and diff == 0:
        diff = 7
    return date + datetime.timedelta(days=diff)


def next_monday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the next Monday.

    :param date: The date from when to date forward.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Monday.
    """
    return next_weekday(date=date, weekday="Mon", exclude_date=exclude_date)


def next_tuesday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the next Tuesday.

    :param date: The date from when to date forward.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Tuesday.
    """
    return next_weekday(date=date, weekday="Tue", exclude_date=exclude_date)


def next_wednesday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the next Wednesday.

    :param date: The date from when to date forward.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Wednesday.
    """
    return next_weekday(date=date, weekday="Wed", exclude_date=exclude_date)


def next_thursday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the next Thursday.

    :param date: The date from when to date forward.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Thursday.
    """
    return next_weekday(date=date, weekday="Thu", exclude_date=exclude_date)


def next_friday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the next Friday.

    :param date: The date from when to date forward.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Friday.
    """
    return next_weekday(date=date, weekday="Fri", exclude_date=exclude_date)


def next_saturday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the next Saturday.

    :param date: The date from when to date forward.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Saturday.
    """
    return next_weekday(date=date, weekday="Sat", exclude_date=exclude_date)


def next_sunday(date: datetime.date, exclude_date: bool = False) -> datetime.date:
    """Get the date of the next Sunday.

    :param date: The date from when to date forward.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Sunday.
    """
    return next_weekday(date=date, weekday="Sun", exclude_date=exclude_date)


def this_weekday(date: datetime.date, weekday: int | str) -> datetime.date:
    """Get the date of the specified weekday in the week of date.

    :param date: The date whose week to look for the specified weekday.
    :param weekday: An integer (0 stands for Monday)
        or the name (full or 3-letter abbreviation) of a weekday/weekend.
    :return: The date of the next weekday.
    """
    diff = _format_weekday(weekday) - date.weekday()
    return date + datetime.timedelta(days=diff)


def this_monday(date: datetime.date) -> datetime.date:
    """Get the date of Monday in the week of date.

    :param date: The date whose week to look for Monday.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Monday.
    """
    return this_weekday(date=date, weekday="Mon")


def this_tuesday(date: datetime.date) -> datetime.date:
    """Get the date of Tuesday in the week of date.

    :param date: The date whose week to look for Tuesday.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Tuesday.
    """
    return this_weekday(date=date, weekday="Tue")


def this_wednesday(date: datetime.date) -> datetime.date:
    """Get the date of Wednesday in the week of date.

    :param date: The date whose week to look for Wednesday.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Wednesday.
    """
    return this_weekday(date=date, weekday="Wed")


def this_thursday(date: datetime.date) -> datetime.date:
    """Get the date of Thursday in the week of date.

    :param date: The date whose week to look for Thursday.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Thursday.
    """
    return this_weekday(date=date, weekday="Thu")


def this_friday(date: datetime.date) -> datetime.date:
    """Get the date of Friday in the week of date.

    :param date: The date whose week to look for Friday.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Friday.
    """
    return this_weekday(date=date, weekday="Fri")


def this_saturday(date: datetime.date) -> datetime.date:
    """Get the date of Saturday in the week of date.

    :param date: The date whose week to look for Saturday.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Saturday.
    """
    return this_weekday(date=date, weekday="Sat")


def this_sunday(date: datetime.date) -> datetime.date:
    """Get the date of Sunday in the week of date.

    :param date: The date whose week to look for Sunday.
    :param exclude_date: Whether to exclude `date` when dating forward.
    :return: The date of the next Sunday.
    """
    return this_weekday(date=date, weekday="Sun")



"""Update shebang of Python scripts in a directory."""

from pathlib import Path
from argparse import ArgumentParser, Namespace
from magic import Magic


def _update_shebang(path: Path, shebang: str):
    with path.open("r") as fin:
        lines = fin.readlines()
    if lines[0].startswith("#!"):
        if "python" in lines[0]:
            lines[0] = shebang
    with path.open("w") as fout:
        fout.writelines(lines)


def update_shebang(script_dir: Path | str, shebang: str):
    """Update the Shebang of scripts in the given script dir.

    :param script_dir: A directory containing scripts whose shebang are to be updated.
    :param shebang: The new shebang to use.
    """
    if isinstance(script_dir, str):
        script_dir = Path(script_dir)
    if not shebang.startswith("#!"):
        shebang = "#!" + shebang
    if not shebang.endswith("\n"):
        shebang += "\n"
    magic = Magic(mime=True, uncompress=True)
    for path in script_dir.iterdir():
        if path.is_file() and magic.from_file(str(path)).startswith("text/"):
            _update_shebang(path, shebang)


def parse_args(args=None, namespace=None) -> Namespace:
    """Parse command-line arguments.

    :param args: The arguments to parse.
        If None, the arguments from command line is parsed.
    :param namespace: An inital Namespace object.
    :return: A Namespace object containing parsed options.
    """
    parser = ArgumentParser(description="Update shebang of scripts.")
    parser.add_argument(
        "-d",
        "--script-dir",
        dest="script_dir",
        required=True,
        help="The directory containing scripts whose shebang are to be updated.",
    )
    parser.add_argument(
        "-s",
        "--sb",
        "--shebang",
        dest="shebang",
        required=True,
        help="The new shebang to use.",
    )
    return parser.parse_args(args=args, namespace=namespace)


def main():
    """Main function for the module."""
    args = parse_args()
    update_shebang(args.script_dir, args.shebang)


if __name__ == "__main__":
    main()



"""Test the aiutil.logf module."""

from pathlib import Path
import aiutil.hadoop.logf

BASE_DIR = Path(__file__).parent


def test_main():
    output = BASE_DIR / "log_s"
    args = aiutil.hadoop.logf.parse_args(
        [
            "filter",
            str(BASE_DIR / "application_1611634725250_1347938"),
            "-o",
            str(output),
            "-d",
        ]
    )
    aiutil.hadoop.logf.main(args)
    assert output.is_file()
    text = output.read_text()
    assert "- Possible Error Lines -" in text
    assert "- Deduped Error Lines -" in text



"""Hash code related utils."""

import hashlib
from pathlib import Path
from loguru import logger


def rmd5(path: str | Path, output: str | Path = "") -> str:
    """Calculate md5sums recursively for the given path.

    :param path: The path of a file or directory.
    :param output: An optional path to a file to ouput md5sums of files.
    :returns: The md5sum of md5sums of files.
    """
    if isinstance(path, str):
        path = Path(path)
    md5sums = []
    _rmd5(path, md5sums)
    md5sums.sort()
    text = "\n".join(md5sums)
    if output:
        if isinstance(output, str):
            output = Path(output)
        with output.open("w", encoding="utf-8") as fout:
            fout.write(text)
    return hashlib.md5(text.encode()).hexdigest()


def _rmd5(path: Path, res: list[str]) -> None:
    """Helper function of rmd5.

    :param path: The Path object of a file or directory.
    :param res: A list to record the result.
    """
    if path.is_file():
        try:
            md5sum = hashlib.md5(path.read_bytes()).hexdigest()
        except Exception:
            md5sum = "FAILED!"
        line = f"{str(path)}: {md5sum}"
        res.append(line)
        logger.info(line)
        return
    try:
        for p in path.iterdir():
            _rmd5(p, res)
    except Exception:
        line = f"{str(path)}: FAILED!"
        res.append(line)
        logger.info(line)



"""Collections related utills."""

from typing import Callable, Any


def format_item_html(key: Any, value: Any) -> str:
    """Format a (key, value) pair of a dictionary in HTML format.

    :param key: An key of the dictionary.
    :param value: The corresponding value of the key.
    :return: A string representation of the formatted key/value pair in HTML format.
    """
    return "&nbsp;" * 4 + f"{key}: {value}"


def format_item_plain(key: Any, value: Any) -> str:
    """Format a (key, value) pair of a dictionary in HTML format.

    :param key: An key of the dictionary.
    :param value: The corresponding value of the key.
    :return: A string representation of the formatted key/value pair in plain format.
    """
    return " " * 4 + f"{key}: {value}"


def format_dict_html(
    dict_: dict,
    fmt: Callable[[Any, Any], str] = format_item_html,
    filter_: Callable[[Any, Any], bool] = lambda key, value: True,
):
    """Format a dict in HTML format for pretty printing.

    :param dict_: The dictionary to format.
    :param fmt: A function to format a (key, value) pair.
    :param filter_: A filtering function to select items from the dictionary.
    :return: A string representation of the formatted dict in HTML format.
    """
    lines = (fmt(k, v) for k, v in dict_.items() if filter_(k, v))
    return "{<br>" + "<br>".join(lines) + "<br>}"


def format_dict_plain(
    dict_: dict,
    fmt: Callable[[Any, Any], str] = format_item_plain,
    filter_: Callable[[Any, Any], bool] = lambda key, value: True,
) -> str:
    """Format a dict for pretty printing.

    :param dict_: The dictionary to format.
    :param fmt: A function to format a (key, value) pair.
    :param filter_: A filtering function to select items from the dictionary.
    :return: A string representation of the formatted dict in plain format.
    """
    lines = (fmt(k, v) for k, v in dict_.items() if filter_(k, v))
    return "{\n" + "\n".join(lines) + "\n}"



"""DockerHub"""

import requests


class DockerHub:
    """A class for manipulating Docker Hub."""

    def __init__(self, user: str, password: str = "", token: str = ""):
        self.user = user
        self._token = self.token(password) if password else token

    def tags(self, image: str) -> list[str]:
        """Get tags of a Docker image on Docker Hub.

        :param image: The name of a Docker image, e.g., "jupyterhub-ds".
        :return: A list of tags belonging to the Docker image.
        """
        user = self.user
        if "/" in image:
            user, image = image.split("/")
        url = f"https://hub.docker.com/v2/repositories/{user}/{image}/tags/"
        res = requests.get(url, timeout=10)
        return res.json()["results"]

    def token(self, password: str) -> None:
        """Generate a token of the account.

        :param password: The password of the user.
        """
        res = requests.post(
            url="https://hub.docker.com/v2/users/login/",
            data={"username": self.user, "password": password},
            timeout=10,
        )
        self._token = res.json()["token"]

    def delete_tag(self, image: str, tag: str = "") -> str:
        """Delete a tag of the specified Docker image.

        :param image: The name of a docker image (without tag).
        :param tag: The tag of the Docker image (to delete).
        :return: The removed tag of the Docker image.
            An empty string is returned if no tag is removed.
        """
        user = self.user
        if "/" in image:
            user, image = image.split("/")
        if ":" in image:
            image, tag = image.split(":")
        if not tag:
            return ""
        url = f"https://hub.docker.com/v2/repositories/{user}/{image}/tags/{tag}/"
        res = requests.delete(
            url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"JWT {self._token}",
            },
            timeout=10,
        )
        return tag if res else ""
